/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
/*
                          promis / imsil / src
                          --------------------
                                im2dgeo.c

  This file contains the routines for the two-dimensional geometry checks
  for the Monte Carlo Simulator for Ion Implantation in crystalline targets
  IMSIL. The Geometry is assumed to be represented by a finite QUADTREE,
  which is generated by the routines in imquad.c during an initialization
  step.

  AUTHORS:   HS  Hannes Stippel, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HS  1994-02-14  new from scratch
  ===========================================================================*/

#include <stdio.h>
#include "imquad.h"
#include "im2dgeo.h"


/* --------------------------------------------------------------------------*/

/*** g e t N e w R e g i o n
     determines the region of a given point using a quadtree.
     Retrun values: -2    ... an unknowen leaf type was encountered
                    -1    ... point is outside the root rectangle
		     0    ... air region
		     else ... the region number ***/

int getNewRegion (
		  Point      point,	/* testpoint */
		  QuadNode  *quadNode,	/* quadtree root node */
		  QuadLeaf **quadLeaf,	/* quadtree root node */
		  Point      center,	/* center of the root rectangle */
		  Point     *newCenter,	/* center of the rectangle */
		  double     sideX,	/* side x of the root rectangle */
		  double     sideZ,	/* side z of the root rectangle */
		  double    *newSideX,	/* side x of the rectangle */
		  double    *newSideZ,	/* side z of the rectangle */
		  double     pcoord[])	/* coordinates of the points */
{
   int                 iChild, iPoint,		/* counter */
                       iPointFrom,		/* points determining line */
                       iPointTo,		/* in a leaf */
                       numRegions;		/* # of regions for type III */
   QuadLeaf           *quadHelpLeaf;		/* help variable for leaf */
   QuadLeafElementI   *quadLeafElementI;	/* pointer to leaf type I */
   QuadLeafElementII  *quadLeafElementII;	/* pointer to leaf type II */
   QuadLeafElementIII *quadLeafElementIII;	/* pointer to leaf type III */
   
   if (center.x < point.x)
      iChild = (center.z < point.z) ? 3 : 2;
   else
      iChild = (center.z < point.z) ? 1 : 0;
   
   if (IS_LEAF (quadNode, iChild))
   {
      *quadLeaf = quadHelpLeaf = quadNode->quadChild[iChild].quadLeaf;
      switch (quadHelpLeaf->quadLeafType)
      {
         case 1:
	    /* leaf type I */
	    quadLeafElementI =
	       quadHelpLeaf->quadLeafElement.quadLeafElementI;
	    return (quadLeafElementI->region);
	    break;
	 
         case 2:
	    /* leaf type II --> get the region */
	    quadLeafElementII =
	       quadHelpLeaf->quadLeafElement.quadLeafElementII;
	    iPointFrom = 2 * (quadLeafElementII->fromPoint - 1);
	    iPointTo   = 2 * (quadLeafElementII->toPoint   - 1);
	    if ((point.x - pcoord[iPointFrom])              *
		(pcoord[iPointFrom+1] - pcoord[iPointTo+1]) +
		(point.z - pcoord[iPointFrom+1])            *
		(pcoord[iPointTo]     - pcoord[iPointFrom]) > 0.0)
	       /* point is on the right side of the line */
	       return (quadLeafElementII->rightRegion);
	    else
	       /* point is on the left side of the line */
	       return (quadLeafElementII->leftRegion);
	    break;
	 
         case 3:
	    /* leaf type III --> get the region */
	    quadLeafElementIII =
	       quadHelpLeaf->quadLeafElement.quadLeafElementIII;
	    numRegions = quadLeafElementIII->numRegions;
	    iPointFrom = 2 * (quadLeafElementIII->innerPoint - 1);
	    iPointTo   = 2 * (quadLeafElementIII->endPoint[0] - 1);
	    iPoint     = 1;
	    if ((point.x - pcoord[iPointFrom])              *
		(pcoord[iPointFrom+1] - pcoord[iPointTo+1]) +
		(point.z - pcoord[iPointFrom+1])            *
		(pcoord[iPointTo]     - pcoord[iPointFrom]) < 0.0)
	       /* point is on the right side of the line ==> search
		  the first line, where point is on the left side */
	       for (; iPoint < numRegions; iPoint++)
	       {
		  iPointTo = 2 * (quadLeafElementIII->endPoint[iPoint] - 1);
		  if ((point.x - pcoord[iPointFrom])              *
		      (pcoord[iPointFrom+1] - pcoord[iPointTo+1]) +
		      (point.z - pcoord[iPointFrom+1])            *
		      (pcoord[iPointTo]     - pcoord[iPointFrom]) > 0.0)
		     /* point is on the left side of the line */
		     break;
	       }
	 
	    /* search for the first line, where the point is on the
	       right side of the line */
	    for (;iPoint < numRegions; iPoint++)
	    {
	       iPointTo = 2 * (quadLeafElementIII->endPoint[iPoint] - 1);
	       if ((point.x - pcoord[iPointFrom])              *
		   (pcoord[iPointFrom+1] - pcoord[iPointTo+1]) +
		   (point.z - pcoord[iPointFrom+1])            *
		   (pcoord[iPointTo]     - pcoord[iPointFrom]) < 0.0)
		  /* point is on the right side of the line */
		  return (quadLeafElementIII->region[iPoint-1]);
	    }
	    return (quadLeafElementIII->region[numRegions-1]);
	    break;
	 
         default:
	    /* what's that ??? */
	    printf ("\n\n***** ERROR:\n\t");
	    printf ("Unknown leaf type <%d>!\n\n",
		    quadHelpLeaf->quadLeafType);
	    return (-2);
	    break;
      }
   }
   else
   {
      *newSideX = sideX / 2.0;
      *newSideZ = sideZ / 2.0;
      newCenter->x = NEW_X (iChild, center.x, *newSideX / 2.0);
      newCenter->z = NEW_Z (iChild, center.z, *newSideZ / 2.0);
      return (getNewRegion (point, quadNode->quadChild[iChild].quadNode,
			    quadLeaf, *newCenter, newCenter, *newSideX,
			    *newSideZ, newSideX, newSideZ, pcoord));
   }
   
   /* point is outside the root rectangle */
   return (-1);
}

/*** g e t R e g i o n
     determines the region of a given point using a quadtree, taking the
     last region found into account.
     Retrun values: -2    ... an unknowen leaf type was encountered
                    -1    ... point is outside the root rectangle
		     0    ... air region
		     else ... the region number ***/

int getRegion (
	       Point      point,	/* testpoint */
	       QuadNode  *quadRoot,	/* quadtree root node */
	       QuadLeaf  *quadLeaf,	/* last quadtree leaf */
	       QuadLeaf **quadNewLeaf,	/* new quadtree leaf */
	       Point      center,	/* center of the rectangle */
	       Point     *leafCenter,	/* center of the last leaf rectangle */
	       double     sideX,	/* side x of the rectangle */
	       double     sideZ,	/* side z of the rectangle */
	       double    *leafSideX,	/* side x of the last leaf rectangle */
	       double    *leafSideZ,	/* side z of last leaf the rectangle */
	       double     pcoord[])	/* coordinates of the points */
{
   int                 iPoint,			/* counter */
                       iPointFrom,		/* points determining line */
                       iPointTo,		/* in a leaf */
                       numRegions;		/* # of regions for type III */
   QuadLeafElementI   *quadLeafElementI;	/* pointer to leaf type I */
   QuadLeafElementII  *quadLeafElementII;	/* pointer to leaf type II */
   QuadLeafElementIII *quadLeafElementIII;	/* pointer to leaf type III */
   
   /* test, if the point is in the old leaf */
   if (leafCenter->x-*leafSideX/2.0 <= point.x &&
       point.x <= leafCenter->x+*leafSideX/2.0 &&
       leafCenter->z-*leafSideZ/2.0 <= point.z &&
       point.z <= leafCenter->z+*leafSideZ/2.0)
      /* point is in the same leaf! */
      switch (quadLeaf->quadLeafType)
      {
         case 1:
	    /* leaf type I */
	    quadLeafElementI = quadLeaf->quadLeafElement.quadLeafElementI;
	    return (quadLeafElementI->region);
	    break;
	 
         case 2:
	    /* leaf type II --> get the region */
	    quadLeafElementII = quadLeaf->quadLeafElement.quadLeafElementII;
	    iPointFrom = 2 * (quadLeafElementII->fromPoint - 1);
	    iPointTo   = 2 * (quadLeafElementII->toPoint   - 1);
	    if ((point.x - pcoord[iPointFrom])              *
		(pcoord[iPointFrom+1] - pcoord[iPointTo+1]) +
		(point.z - pcoord[iPointFrom+1])            *
		(pcoord[iPointTo]     - pcoord[iPointFrom]) > 0.0)
	       /* point is on the right side of the line */
	       return (quadLeafElementII->rightRegion);
	    else
	       /* point is on the left side of the line */
	       return (quadLeafElementII->leftRegion);
	    break;
	 
         case 3:
	    /* leaf type III --> get the region */
	    quadLeafElementIII = quadLeaf->quadLeafElement.quadLeafElementIII;
	    numRegions = quadLeafElementIII->numRegions;
	    iPointFrom = 2 * (quadLeafElementIII->innerPoint - 1);
	    iPointTo   = 2 * (quadLeafElementIII->endPoint[0] - 1);
	    iPoint     = 1;
	    if ((point.x - pcoord[iPointFrom])              *
		(pcoord[iPointFrom+1] - pcoord[iPointTo+1]) +
		(point.z - pcoord[iPointFrom+1])            *
		(pcoord[iPointTo]     - pcoord[iPointFrom]) < 0.0)
	       /* point is on the right side of the line ==> search
		  the first line, where point is on the left side */
	       for (; iPoint < numRegions; iPoint++)
	       {
		  iPointTo = 2 * (quadLeafElementIII->endPoint[iPoint] - 1);
		  if ((point.x - pcoord[iPointFrom])              *
		      (pcoord[iPointFrom+1] - pcoord[iPointTo+1]) +
		      (point.z - pcoord[iPointFrom+1])            *
		      (pcoord[iPointTo]     - pcoord[iPointFrom]) > 0.0)
		     /* point is on the left side of the line */
		     break;
	       }
	 
	    /* search for the first line, where the point is on the
	       right side of the line */
	    for (; iPoint < numRegions; iPoint++)
	    {
	       iPointTo = 2 * (quadLeafElementIII->endPoint[iPoint] - 1);
	       if ((point.x - pcoord[iPointFrom])              *
		   (pcoord[iPointFrom+1] - pcoord[iPointTo+1]) +
		   (point.z - pcoord[iPointFrom+1])            *
		   (pcoord[iPointTo]     - pcoord[iPointFrom]) < 0.0)
		  /* point is on the right side of the line */
		  return (quadLeafElementIII->region[iPoint-1]);
	    }
	    return (quadLeafElementIII->region[numRegions-1]);
	    break;
	 
         default:
	    /* what's that ??? */
	    printf ("\n\n***** ERROR:\n\t");
	    printf ("Unknown leaf type <%d>!\n\n", quadLeaf->quadLeafType);
	    return (-2);
	    break;
      }
   else
      return (getNewRegion (point, quadRoot, quadNewLeaf, center, leafCenter,
			    sideX, sideZ, leafSideX, leafSideZ, pcoord));
}

/*** g e t E n t r y P o i n t
     computes the entrance point into the geometry. Sorry, but this is done
     by the slow intersection method. Can't think of anything faster though
     at the moment.
     Retrun values: -4    ... zero motion
                    -3    ... no intersection with the border found
                    -2    ... an unknowen leaf type was encountered
                    -1    ... intersection point is outside the root rectangle
		     0    ... air region
		     else ... the region number ***/

int getEntryPoint (
		   Point      point,		/* location of ion */
		   double     directionX,	/* direction of motion x */
		   double     directionZ,	/* direction of motion z */
		   int        numAmbients,	/* # of lines for the hull */
		   int        ambientBorder[],	/* lines of simulation hull */
		   double     pcoord[],		/* coordinates of the points */
		   QuadNode  *quadRoot,		/* root node of the quadtree */
		   QuadLeaf **quadLeaf,		/* quadtree root node */
		   Point      center,		/* center of root rectangle */
		   Point     *newCenter,	/* center of leaf rectangle */
		   double     rootSideX,	/* side x of root rectangle */
		   double     rootSideZ,	/* side z of root rectangle */
		   double    *newSideX,		/* side x of leaf rectangle */
		   double    *newSideZ,		/* side z of leaf rectangle */
		   double     overshoot,	/* entrance length to area */
		   Point     *entry)		/* entry point */
{
   int    iPointFrom, iPointTo,	/* indices of points defining the line */
          iAmbient,		/* counter */
          firstMue = 0;		/* flag, if a mue has already been found */
   double vectorX, vectorZ,	/* vector of the line */
          determinante,		/* determinante */
          lambda, mue,		/* intersection point */
          minMue;		/* nearest intersection point */
   Point  testPoint;		/* for testing the entered region */
   
   iPointFrom = 2 * (ambientBorder[numAmbients-1] - 1);
   
   for (iAmbient = 0; iAmbient < numAmbients; iAmbient++)
   {
      /* get the end point of the line */
      iPointTo = 2 * (ambientBorder[iAmbient] - 1);
      
      /* determine the vector of the border line */
      vectorX = pcoord[iPointTo]   - pcoord[iPointFrom];
      vectorZ = pcoord[iPointTo+1] - pcoord[iPointFrom+1];
      
      /* the determinante */
      determinante = vectorZ * directionX - vectorX * directionZ;
      if (determinante != 0.0)
      {
	 /* compute the intersection point in the coordinate system of the
	    border line */
	 lambda = (directionX * (point.z - pcoord[iPointFrom+1]) -
		   directionZ * (point.x - pcoord[iPointFrom]))  /
		      determinante;
	 if (0.0 <= lambda && lambda < 1.0)
	    /* intersection is possible */
	 {
	    if (directionX != 0.0)
	       mue = (pcoord[iPointFrom] - point.x + lambda * vectorX) /
		  directionX;
	    else
	       if (directionZ != 0.0)
		  mue = (pcoord[iPointFrom+1] - point.z + lambda * vectorZ) /
		     directionZ;
	       else
		  /* zero motion ??? */
		  return (-4);
	    
	    if (mue >= 0.0)
	       /* this is an intersection */
	       if (firstMue == 0)
	       {
		  minMue   = mue;
		  firstMue = 1;
	       }
	       else
		  if (mue < minMue)
		     minMue = mue;
	 }
      }
      
      /* set the index for the next beginning point of the next line */
      iPointFrom = iPointTo;
   }
   
   if (firstMue == 0)
      /* no intersection found */
      return (-3);
   
   /* compute the entrance point of the ion */
   entry->x = point.x + minMue * directionX;
   entry->z = point.z + minMue * directionZ;
   
   /* determine the region which is entered */
   testPoint.x = point.x + minMue * (1.0 + overshoot) * directionX;
   testPoint.z = point.z + minMue * (1.0 + overshoot) * directionZ;
   return (getNewRegion (testPoint, quadRoot, quadLeaf, center, newCenter,
			 rootSideX, rootSideZ, newSideX, newSideZ, pcoord));
}
