/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
/*
                          promis / imsil / src
                          --------------------
                                 imf2c.c
 
  This file contains the FORTRAN/C interface for the routines needed to
  discretize a given given geometry by a finite QUADTREE and to perform the
  necessary geometry checks by the Monte Carlo Simulator for Ion Implantation
  in crystalline targets IMSIL.

  AUTHORS:   HS  Hannes Stippel, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HS  1994-02-01  new from scratch
  HS  1994-11-05  added graphics interface
  HW  1994-11-07  added trnini,trxz
  ===========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "xpgp.h"
#include "imquad.h"
#include "imgetmem.h"
#include "imtime.h"
#include "im2dgeo.h"
#include "imtestquad.h"
#include "imvisualquad.h"
#include "imsurfgeo.h"

/* --------------------------------------------------------------------------*/


/* ================================
   local defines (for the graphics)
   ================================ */

#define BORDER_X 20
#define BORDER_Z 20
#define SIZE_X 740
#define SIZE_Z 540

#if defined(__hpux) || defined(vms) || defined(__vms)
#else
#define F_Move f_move_ 
#define F_LineTo f_lineto_ 
#endif

/* --------------------------------------------------------------------------*/


/* ================
   global variables
   ================ */

/* the quad tree */
QuadNode *quadTree;	/* root of quadtree */
QuadLeaf *quadLeaf;	/* last found leaf rectangle */
Point     rootCenter,	/* center of the root rectangle */
          leafCenter;	/* center of last found leaf */
double    rootSideX,	/* side length of the root rectangle x */
          rootSideZ,	/* side length of the root rectangle z */
          newSideX,	/* side length of last leaf rectangle x */
          newSideZ,	/* side length of last leaf rectangle z */
          overshoot,	/* distance of entry into new region */
	 *ambientBorderLength; /*Length of Ambients */
int       numAmbients,	/* number of lines defining the simulation area */
         *ambientBorder;/* lines of the border of the simulation area */
int 	  istat;	/* status flag for F_graphics calls */

/* the visualization */
double offsetX,	/* x-offset to fit geometry into window */
       offsetZ,	/* z-offset to fit geometry into window */
       scaleX,	/* scaling factor x to fit geometry into window */
       scaleZ;	/* scaling factor z to fit geometry into window */

double *transform; /* transformation matrix , see TRNINI for definition */

/* --------------------------------------------------------------------------*/

/* external global variables
   ========================= */

	/* --------------------------------------------------------------------------*/

     /*** i m q u a d
     generates the quadtree and the surrounding border.
     Error codes: 0   ... everything is OK
                  1-7 ... not enough memory
		  8   ... error during generating the additional references
		  9   ... error during the generation of the quadtree
		  10  ... error during generating the ambient border ***/


#if defined(__hpux) || defined(vms) || defined(__vms)
void imquad  (
#else
void imquad_ (
#endif


	      long   *numFaces,		/* number of faces */
	      long   *faceEndIndex,	/* end indices of the faces */
	      long   *lineIndex,	/* point indices defining the faces */
	      long   *numPoints,	/* number of points */
	      double *pcoord,		/* coordinates of the points */
	      double *epsma,		/* machine accuracy */
	      double *xscale,		/* quater of the grid constante */
	      long   *iError)		/* error variable */
{
   LineType *pointIndex;		/* lines defined over the point indices */
   int     numLines,		/* number of lines */
          *borderLine,		/* line is border between which regions */
          *pointEndIndex,	/* end indices for pointIsInLine array */
          *pointIsInLine,	/* point is in which lines */
           numCuts,		/* number of intersecting lines */
          *cut,			/* intersection information */
           numInsides,		/* number of points inside a rectangle */
          *inside,		/* points inside the rectangle */
           iLine,		/* line counter */
           iPoint,		/* point counter */
           iCut,		/* number of intersecting lines */
           iInside;		/* number of points inside a rectangle */
   double  stime,		/* time variable */
/*           minX, maxX,	 minimum and maximum of simulation area x */
/*           minZ, maxZ , 	minimum and maximum of simulation area z */
           quadResolutionX,	/* reolution of quadtree in x direction */
           quadResolutionZ;	/* reolution of quadtree in z direction */
   
   /* hopefully everything goes OK */
   *iError = 0;
   
   /* this is the maximum possible number of lines */
   numLines = faceEndIndex[*numFaces-1];
   
   /* allocate memory */
   if ((pointIndex = (LineType *) malloc (sizeof (LineType) * numLines)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for lines!");
      printf ("\n\n");
      *iError = 1;
      return;
   }
   if ((borderLine = (int *) malloc (sizeof (int) * numLines * 2)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for ");
      printf ("borderLine!\n\n");
      *iError = 2;
      return;
   }
   if ((pointEndIndex = (int *) malloc (sizeof (int) * *numPoints)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for ");
      printf ("pointEndIndex!\n\n");
      *iError = 3;
      return;
   }
   if ((pointIsInLine = (int *) malloc (sizeof (int) * numLines * 2)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for ");
      printf ("pointIsInLine!\n\n");
      *iError = 4;
      return;
   }
      
   /* generate the required references */
   stime = imtime ();
   quadMakeReferences (*numFaces, faceEndIndex, lineIndex, &numLines,
		       pointIndex, borderLine, *numPoints, pointEndIndex,
		       pointIsInLine, pcoord);
   stime -= imtime ();
#ifdef _MEASURE_TIME_
   printf ("\n\n***** INFO:\n\tMakeReferences time: %g\n\n", fabs(stime));
#endif
   
   /* allocate memory */
   if ((cut = (int *) malloc (sizeof (int) * numLines)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for cuts!\n\n");
      *iError = 5;
      return;
   }
   if ((inside = (int *) malloc (sizeof (int) * numLines)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for inside!");
      printf ("\n\n");
      *iError = 6;
      return;
   }
   
   /* initialize the information about intersecting lines and points inside
      the rectangles */
   numCuts = numLines;
   for (iCut = 0; iCut < numCuts; iCut++)
      cut[iCut] = iCut + 1;
   numInsides = *numPoints;
   for (iInside = 0; iInside < numInsides; iInside++)
      inside[iInside] = iInside + 1;
   
   /* determine the root rectangle */
   minX = maxX = pcoord[0];
   minZ = maxZ = pcoord[1];
   for (iPoint = 1; iPoint < *numPoints; iPoint++)
   {
      if (pcoord[2*iPoint]   < minX) minX = pcoord[2*iPoint];
      if (pcoord[2*iPoint]   > maxX) maxX = pcoord[2*iPoint];
      if (pcoord[2*iPoint+1] < minZ) minZ = pcoord[2*iPoint+1];
      if (pcoord[2*iPoint+1] > maxZ) maxZ = pcoord[2*iPoint+1];
   }
   rootCenter.x = (minX + maxX) / 2.0;
   rootCenter.z = (minZ + maxZ) / 2.0;
   rootSideX = maxX - minX;
   rootSideZ = maxZ - minZ;
   
   /* set the minimum resolution */
   quadResolutionX = 1.0e3 * rootSideX * *epsma;
   quadResolutionZ = 1.0e3 * rootSideZ * *epsma;
   
   /* gnerate the quadtree */
   stime = imtime ();
   if ((quadTree = quadMakeTree (*numFaces, faceEndIndex, lineIndex,
				 numLines, pointIndex, borderLine, numCuts,
				 cut, *numPoints, pointEndIndex,
				 pointIsInLine, pcoord, numInsides, inside,
				 rootCenter, rootSideX, rootSideZ,
				 quadResolutionX, quadResolutionZ)) ==
       NODE_NULL)
   {
      printf ("\n\n ***** ERROR:\n\t");
      printf ("Occured during initialization of the QuadTree!\n\n");
      *iError = 9;
      return;
   }
   stime -= imtime ();
#ifdef _MEASURE_TIME_
   printf ("\n\n***** INFO:\n\tQuadTree time: %g\n\n", fabs(stime));
#endif
#ifdef _MEM_STATISTICS_
   printf ("\n\n***** INFO:\n");
   printf ("\tNumber of Allocations for Nodes: %d\n\t\t", numNodeAllocs);
   printf ("Number of Nodes: %d\n", numNodes);
   printf ("\tNumber of Allocations for Leafs: %d\n\t\t", numLeafAllocs);
   printf ("Number of Leafs: %d\n", numLeafs);
   printf ("\tNumber of Allocations for ElementIs: %d\n\t\t",
	   numLeafElementIAllocs);
   printf ("Number of ElementIs: %d\n", numLeafElementIs);
   printf ("\tNumber of Allocations for ElementIIs: %d\n\t\t",
	   numLeafElementIIAllocs);
   printf ("Number of ElementIIs: %d\n", numLeafElementIIs);
   printf ("\tNumber of Allocations for ElementIIIs: %d\n\t\t",
	   numLeafElementIIIAllocs);
   printf ("Number of ElementIIIs: %d\n", numLeafElementIIIs);
   printf ("\tNumber of Allocations for Cuts: %d\n\t\t", numCutAllocs);
   printf ("Number of Cuts: %d\n", numCuts);
   printf ("\tNumber of Allocations for Insides: %d\n\t\t", numInsideAllocs);
   printf ("Number of Insides: %d\n", numInsides);
   printf ("\tNumber of Allocations for ElementIIIAdds: %d\n\t\t",
	   numLeafElementIIIAddAllocs);
   printf ("Number of ElementIIIAdds: %d\n", numLeafElementIIIAdds);
#endif
   
   /* get the number of lines forming the boundary of the simulation area */
   numAmbients = 0;
   for (iLine = 0; iLine < numLines; iLine++)
      if (borderLine[2*iLine] == 0 || borderLine[2*iLine+1] == 0)
      {
	 /* found a new piece of the border */
	 numAmbients++;
      }

   /* allocate memory */
   if ((ambientBorder = (int *) malloc (sizeof (int) * numAmbients)) == NULL) 
	
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for ");
      printf ("Ambient!\n\n");
      *iError = 7;
      return;
   }
   
   stime = imtime ();

   if (quadMakeAmbient (numLines, pointIndex, borderLine, numAmbients,
		     ambientBorder) != 0)
   {
      printf ("\n\n ***** ERROR:\n\t");
      printf ("Occured during the generation of the border of the "); 
      printf ("simulation area!\n\n");
      *iError = 10;
      return;
   }

   imsurfMake (numAmbients,ambientBorder,pcoord,minX,maxX,minZ,maxZ);


   stime -= imtime ();

#ifdef _MEASURE_TIME_
   printf ("\n\n***** INFO:\n\tAmbient generation time: %g\n\n", fabs(stime));
#endif
   
   /* compute the overshoot for the determination of the region where an ion
      (re-)enters the target */
   overshoot = fabs (pcoord[0] - pcoord[2]);
   for (iPoint = 2; iPoint < 2 * (*numPoints - 1); iPoint++)
      if (fabs (pcoord[iPoint-1] - pcoord[iPoint+1]) < overshoot ||
	  overshoot <= 0.0)
	 overshoot = fabs (pcoord[iPoint-1] - pcoord[iPoint+1]);
   overshoot *= 1.0e3 * *epsma;
/* Correction(?) 12-Sep-94 ob sinnvoll oder nicht weiss ich nicht (HW)*/
   overshoot = 2*(quadResolutionX+quadResolutionZ);
#ifdef _DEBUG_QUADTREE_
   testQuadMain (quadTree, *numFaces, faceEndIndex, lineIndex, numLines,
		 pointIndex, borderLine, *numPoints, pointEndIndex,
		 pointIsInLine, pcoord, rootSideX, rootSideZ,
		 quadResolutionX, quadResolutionZ, *epsma, minX, maxX, minZ,
		 maxZ, rootCenter, numAmbients, ambientBorder, overshoot);
#endif
   
   /* free the allocated and not anymore needed memory */
   free (pointIndex);
   free (borderLine);
   free (pointEndIndex);
   free (pointIsInLine);

#ifdef _VISUALIZE_QUADTREE_
   if (visualQuadMain (quadTree, rootCenter, rootSideX, rootSideZ,
		       minX, maxX, minZ, maxZ, *numFaces, pcoord, *xscale)
       != 0)
   {
      printf ("\n\n ***** ERROR:\n\t");
      printf ("Occured during the visualization of the quadtree!\n\n");
      return;
   }
#endif
}

/*** i m g r e g
     determines the region in which a given point lies.
     Region values: -1    ... point is outside the root rectangle
                     0    ... point is in the air region
		     > 0  ... region number
		     else ... error
     Error codes:    0    ... everything is OK
                     else ... error ***/

#if defined(__hpux) || defined(vms) || defined (__vms)
void imgreg  (
#else
void imgreg_ (
#endif

	      
	      long   *useOldLeaf,	/* use old leaf or not */
	      double *locationX,	/* x coordinate of the test point */
	      double *locationZ,	/* z coordinate of the test point */
	      double *pcoord,		/* coordinates of the points */
	      long   *region,		/* region in which the point lies */
	      long   *iError)		/* error code */
{
   double lX, lZ;	/* sidelength */
   Point point;		/* testpoint */
   
   /* hopefully no error */
   *iError = 0;
   
   /* test, if the point is outside the root rectangle */
   lX = rootSideX / 2.0;
   lZ = rootSideZ / 2.0;
   if (rootCenter.x-lX > *locationX || rootCenter.x+lX < *locationX ||
       rootCenter.z-lZ > *locationZ || rootCenter.z+lZ < *locationZ)
   {
      *region = -1;
      return;
   }
   
   /* set the point */
   point.x = *locationX;
   point.z = *locationZ;
   
   if (*useOldLeaf == 1)
   {
      /* use the old leaf */
      *region = getRegion (point, quadTree, quadLeaf, &quadLeaf, rootCenter,
			   &leafCenter, rootSideX, rootSideZ, &newSideX,
			   &newSideZ, pcoord);
      if (*region < -1)
	 /* error */
	 *iError = 1;
   }
   else
   {
      /* do not use the old leaf */
      *region = getNewRegion (point, quadTree, &quadLeaf, rootCenter,
			      &leafCenter, rootSideX, rootSideZ, &newSideX,
			      &newSideZ, pcoord);
      if (*region < -1)
	 /* error */
	 *iError = 1;
   }
}

/*** i m e n t r
     determines the entrance point into the simulation area.
     Region values: -1    ... intersection point is outside the root rectangle
                     0    ... intersection point is in the air region
		     > 0  ... region number
                     else ... error
     Error codes:    0    ... everything is OK
                     else ... error ***/

#if defined(__hpux) || defined(vms) || defined(__vms)
void imentr  (
#else
void imentr_ (
#endif
	      
	      double *locationX,	/* x coordinate of the test point */
	      double *locationZ,	/* z coordinate of the test point */
	      double *directionX,	/* x component of the motion */
	      double *directionZ,	/* z component of the motion */
	      double *pcoord,		/* coordinates of the points */
	      long   *intersection,	/* 1, if an intersection was found */
	      double *newLocationX,	/* x coordinate of the entry test */
	      double *newLocationZ,	/* z coordinate of the entry test */
	      long   *region,		/* region in which the point lies */
	      long   *iError)		/* error code */
{
   Point point,		/* test point */
         entry;		/* entry point into the simulation area */
   
   /* hopefully no error */
   *iError = 0;
   *intersection = 1;
   
   /* set the test point */
   point.x = *locationX;
   point.z = *locationZ;
   
   /* determine the entrance point */
   *region = getEntryPoint (point, *directionX, *directionZ, numAmbients,
			    ambientBorder, pcoord, quadTree, &quadLeaf,
			    rootCenter, &leafCenter, rootSideX, rootSideZ,
			    &newSideX, &newSideZ, overshoot, &entry);
   if (*region == -2 || *region == -4)
      /* error */
      *iError = 1;
   else
      if (*region == -3)
	 /* no intersection */
	 *intersection = 0;
      else
	 if (*region == 0)
	 {
	    *region = getNewRegion (entry, quadTree, &quadLeaf, rootCenter,
				    &leafCenter, rootSideX, rootSideZ,
				    &newSideX, &newSideZ, pcoord);
	    if (*region < -1)
	       /* error */
	       *iError = 1;
	    if (*region == 0)
	       /* no intersection */
	       *intersection = 0;
	 }
   
   /* set the entry point */
   *newLocationX = entry.x;
   *newLocationZ = entry.z;
}


/* Horst Wagner */
/*** t r n i n i
     initialize offset and scaling factors. ***/

#if defined(__hpux) || defined(vms) || defined(__vms)
void trnini  (
#else
void trnini_ (
#endif
  
      	      double *transmat)         /* transformation matrix */
{

transform = transmat;   

/* transformation matrix defined as follows : 
   XSCREEN = XX*(X-XSH)+XY*(Y-YSH)+XZ*(Z-ZSH)
   YSCREEN = YX*(X-XSH)+YY*(Y-YSH)+YZ*(Z-ZSH)

   SCREEN#1 : XX = TRANSMAT(0)
              XY = TRANSMAT(1)
              .....
              ZSH = TRANSMAT(8)
   SCREEN#2 : XX = TRANSMAT(9)
              ....                                */

}

/* Horst Wagner */
/*** t r n x y z
     Transform simulation area coordinates to pixel coordinates ***/

#if defined(__hpux) || defined(vms) || defined(__vms)
void trnxyz  (
#else
void trnxyz_ (
#endif

      	      int *iloc,		/* locout number */
	      double *x,		/* x coordinate in A */
      	      double *y,		/* y coordinate in A */
	      double *z,		/* z coordinate in A */
	      double *xx,			/* x coordinate in pixel*/
	      double *xy)			/* y coordinate in pixel*/
{

double *a;

a = transform + 9*(*iloc-1)-1;
*xx = (*x-a[7])*a[1] + (*y-a[8])*a[2] + (*z-a[9])*a[3];
*xy = (*x-a[7])*a[4] + (*y-a[8])*a[5] + (*z-a[9])*a[6];

}

/* Horst Wagner */
/*** t r n z y x
     Transform pixel ccordinates simulation area coordinates ***/

#if defined(__hpux) || defined(vms) || defined(__vms)
void trnzyx  (
#else
void trnzyx_ (
#endif

      	      int *iloc,		/* locout number */
	      double *x,		/* x coordinate in A */
      	      double *y,		/* y coordinate in A */
	      double *z,		/* z coordinate in A */
	      double *xx,			/* x coordinate in pixel*/
	      double *xy)			/* y coordinate in pixel*/
{

double *a;

a = transform + 9*(*iloc-1)-1;
*x = *xx/a[1]+a[7];
*y = 0;
*z = *xy/a[6]+a[9];
}

#if defined(__hpux) || defined(vms) || defined(__vms)
#else
#define trnxyz trnxyz_ 
#endif
/* define needed for call of trnxyz by drwgeo */


/*** d r w g e o
     draws the geometry in the window. ***/

#if defined(__hpux) || defined(vms) || defined(__vms)
void drwgeo  (
#else
void drwgeo_ (
#endif
      	      int    *iloc,		/* location output # */
	      long   *numFaces,		/* number of faces */
	      long   *faceEndIndex,	/* end indices of the faces */
	      long   *lineIndex,	/* point indices defining the faces */
	      double *pcoord,		/* coordinates of the points */
	      long   *lquad)		/* determines quadtree visualization */
{
   int    iFace,	/* counter for faces */
          iBegin,	/* index to first line of the face */
          iLine,	/* counter for lines */
          iPoint;	/* counter for points */
   double PosX,PosX2;  	/* x-position of a point */
   double PosZ,PosZ2; 	/* z-position of a point */
   double Y;            /* dummy Y -coordinate */   
        
   /* 2D geometry */

   Y = 0;

   /* begin with the first line */
   iBegin = 0;
   SetLineWidth(2);

   /* loop over all faces */
   for (iFace = 0; iFace < *numFaces; iFace++)
   {
      /* the starting point for the first line */
      iPoint = lineIndex[faceEndIndex[iFace]-1] - 1;

      /* determine the coordinates in the plot window */
      trnxyz(iloc,pcoord+2*iPoint,&Y,pcoord+2*iPoint+1,&PosX,&PosZ);
	
      /* move to the starting point */
      Move(PosX,PosZ);
      
      /* loop over all lines of a face */
      for (iLine = iBegin; iLine < faceEndIndex[iFace]; iLine++)
      {
	 iPoint = lineIndex[iLine] - 1;
	 
	 /* determine the coordinates in the plot window */
         trnxyz(iloc,pcoord+2*iPoint,&Y,pcoord+2*iPoint+1,&PosX,&PosZ);
	 
	 /* draw the line */
    	 LineTo(PosX,PosZ);      
      }
      
      iBegin = faceEndIndex[iFace];
   }
   
   /* set the linewidth for the trajectories and the quadtree */
   SetLineWidth(1.); 

   if (*lquad == 1)
   {
      /* set the color for the quadtree (blue) */
      Pen(3);
      /* draw the border of the quadtree */

      trnxyz(iloc,&minX,&Y,&minZ,&PosX,&PosZ);
      trnxyz(iloc,&maxX,&Y,&maxZ,&PosX2,&PosZ2);
      RectangleS(PosX,PosZ,PosX2,PosZ2);
            
      /* draw the quadtree */
      drawQuadTree (quadTree, rootCenter, rootSideX, rootSideZ, iloc);
   }
   
   /* set the color for the trajectories (black) */
   Pen(1);

}

