/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
/*
                          promis / imsil / src
                          --------------------
                               imgetmem.c

  This file contains the routines to allocate the required memory for the
  QUADTREE representation of the geometry.

  AUTHORS:   HS  Hannes Stippel, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HS  1994-02-01  new from scratch
  ===========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "imquad.h"
#include "imgetmem.h"

/* --------------------------------------------------------------------------*/

#ifdef _MEM_STATISTICS_
/* === These are the global counter for the memory statistics === */
int numNodes;			/* number of allocated nodes */
int numNodeAllocs;		/* number of allocations for the nodes */
int numLeafs;			/* number of allocated leafs */
int numLeafAllocs;		/* number of allocations for the leafs */
int numLeafElementIs;		/* number of allocated elements of Type I */
int numLeafElementIAllocs;	/* number of allocations for the element I */
int numLeafElementIIs;		/* number of allocated elements of Type II */
int numLeafElementIIAllocs;	/* number of allocations for the element II */
int numLeafElementIIIs;		/* number of allocated elements of Type III */
int numLeafElementIIIAllocs;	/* number of allocations for the element III */
int numCuts;			/* number of allocated cutinformations */
int numCutAllocs;		/* number of allocations for cutinformation */
int numInsides;			/* number of allocated inside informations */
int numInsideAllocs;		/* number of allocations for inside points */
int numLeafElementIIIAddAllocs;	/* # of additional allocs for element III */
int numLeafElementIIIAdds;	/* # of additional alloc elements III */
#endif

/* --------------------------------------------------------------------------*/

/*** q u a d N o d e A l l o c
     allocates memory for a certain number (NUM_NODE_ALLOC, see imgetmem.h)
     of nodes of a finite quadtree ***/

QuadNode *quadNodeAlloc ()
{
   static int       position = NUM_NODE_ALLOC;	/* position counter in the
						   allocated node array */
   static QuadNode *quadNode;			/* next free node */
   
#ifdef _MEM_STATISTICS_
   /* for the statistics */
   numNodes++;
#endif
   
   /* increment position counter */
   position++;
   
   /* test if new memory allocation is necessary */
   if (position >= NUM_NODE_ALLOC)
   {
      /* allocate memory for next NUM_NODE_ALLOC nodes */
      if ((quadNode = (QuadNode *) malloc (NUM_NODE_ALLOC * sizeof
					   (QuadNode))) == NULL)
      {
	 /* memory allocation failed */
	 printf ("\n\n***** ERROR:\n\tCould not allocate space for Node!\n\n");
	 return (NODE_NULL);
      }
      position = 0;
#ifdef _MEM_STATISTICS_
      /* for the statistics */
      numNodeAllocs++;
#endif
   }
   else
      /* increment node pointer */
      quadNode++;
   
   return (quadNode);
}

/*** q u a d L e a f A l l o c
     allocates memory for a certain number (NUM_LEAF_ALLOC, see imgetmem.h)
     of leafs of an quadtree ***/

QuadLeaf *quadLeafAlloc ()
{
   static int       position = NUM_LEAF_ALLOC;	/* position counter in the
						   allocated node array */
   static QuadLeaf *quadLeaf;			/* next free leaf */
   
#ifdef _MEM_STATISTICS_
   /* for the statistics */
   numLeafs++;
#endif
   
   /* increment position counter */
   position++;
   
   /* test if new memory allocation is necessary */
   if (position >= NUM_LEAF_ALLOC)
   {
      /* allocate memory for next NUM_LEAF_ALLOC nodes */
      if ((quadLeaf = (QuadLeaf *) malloc (NUM_LEAF_ALLOC * sizeof
					   (QuadLeaf))) == NULL)
      {
	 /* memory allocation failed */
	 printf ("\n\n***** ERROR:\n\tCould not allocate space for Leaf!\n\n");
	 return (LEAF_NULL);
      }
      position = 0;
#ifdef _MEM_STATISTICS_
      /* for the statistics */
      numLeafAllocs++;
#endif
   }
   else
      /* increment node pointer */
      quadLeaf++;
   
   return (quadLeaf);
}

/*** q u a d L e a f E l e m e n t I A l l o c
     allocates memory for a certain number (NUM_LEAF_ELEMENT_I_ALLOC, see
     imgetmem.h) of type I leafelements of an quadtree ***/

QuadLeafElementI *quadLeafElementIAlloc ()
{
   static int       position = NUM_LEAF_ELEMENT_I_ALLOC;
                    /* position counter in the allocated node array */
   static QuadLeafElementI *quadLeafElementI;	/* next free leaf */
   
#ifdef _MEM_STATISTICS_
   /* for the statistics */
   numLeafElementIs++;
#endif
   
   /* increment position counter */
   position++;
   
   /* test if new memory allocation is necessary */
   if (position >= NUM_LEAF_ELEMENT_I_ALLOC)
   {
      /* allocate memory for next NUM_LEAF_ELEMENT_I_ALLOC nodes */
      if ((quadLeafElementI = (QuadLeafElementI *) malloc
	   (NUM_LEAF_ELEMENT_I_ALLOC * sizeof (QuadLeafElementI))) == NULL)
      {
	 /* memory allocation failed */
	 printf ("\n\n**** ERROR:\n\tCould not allocate space for ");
	 printf ("LeafElementI!\n\n"); 
	 return (LEAF_ELEMENT_I_NULL);
      }
      position = 0;
#ifdef _MEM_STATISTICS_
      /* for the statistics */
      numLeafElementIAllocs++;
#endif
   }
   else
      /* increment node pointer */
      quadLeafElementI++;
   
   return (quadLeafElementI);
}

/*** q u a d L e a f E l e m e n t I A l l o c
     allocates memory for a certain number (NUM_LEAF_ELEMENT_II_ALLOC, see
     imgetmem.h) of type I leafelements of an quadtree ***/

QuadLeafElementII *quadLeafElementIIAlloc ()
{
   static int       position = NUM_LEAF_ELEMENT_II_ALLOC;
                    /* position counter in the allocated node array */
   static QuadLeafElementII *quadLeafElementII;	/* next free leaf */
   
#ifdef _MEM_STATISTICS_
   /* for the statistics */
   numLeafElementIIs++;
#endif
   
   /* increment position counter */
   position++;
   
   /* test if new memory allocation is necessary */
   if (position >= NUM_LEAF_ELEMENT_II_ALLOC)
   {
      /* allocate memory for next NUM_LEAF_ELEMENT_II_ALLOC nodes */
      if ((quadLeafElementII = (QuadLeafElementII *) malloc
	   (NUM_LEAF_ELEMENT_II_ALLOC * sizeof (QuadLeafElementII))) == NULL)
      {
	 /* memory allocation failed */
	 printf ("\n\n***** ERROR:\n\tCould not allocate space for ");
	 printf ("LeafElementII!\n\n");
	 return (LEAF_ELEMENT_II_NULL);
      }
      position = 0;
#ifdef _MEM_STATISTICS_
      /* for the statistics */
      numLeafElementIIAllocs++;
#endif
   }
   else
      /* increment node pointer */
      quadLeafElementII++;
   
   return (quadLeafElementII);
}

/*** q u a d L e a f E l e m e n t I A l l o c
     allocates memory for a certain number (NUM_LEAF_ELEMENT_III_ALLOC, see
     imgetmem.h) of type I leafelements of an quadtree ***/

QuadLeafElementIII *quadLeafElementIIIAlloc ()
{
   static int       position = NUM_LEAF_ELEMENT_III_ALLOC;
                    /* position counter in the allocated node array */
   static QuadLeafElementIII *quadLeafElementIII;	/* next free leaf */
   
#ifdef _MEM_STATISTICS_
   /* for the statistics */
   numLeafElementIIIs++;
#endif
   
   /* increment position counter */
   position++;
   
   /* test if new memory allocation is necessary */
   if (position >= NUM_LEAF_ELEMENT_III_ALLOC)
   {
      /* allocate memory for next NUM_LEAF_ELEMENT_III_ALLOC nodes */
      if ((quadLeafElementIII = (QuadLeafElementIII *) malloc
	   (NUM_LEAF_ELEMENT_III_ALLOC * sizeof (QuadLeafElementIII))) == NULL)
      {
	 /* memory allocation failed */
	 printf ("\n\n***** ERROR:\n\tCould not allocate space for ");
	 printf ("LeafElementIII!\n\n");
	 return (LEAF_ELEMENT_III_NULL);
      }
      position = 0;
#ifdef _MEM_STATISTICS_
      /* for the statistics */
      numLeafElementIIIAllocs++;
#endif
   }
   else
      /* increment node pointer */
      quadLeafElementIII++;
   
   return (quadLeafElementIII);
}

/*** q u a d C u t A l l o c
     allocates memory for a certain number (NUM_CUT_ALLOC, see imgetmem.h)
     of integers for entries of lines ***/

int *quadCutAlloc (int numAlloc)
{
   static int  position = NUM_CUT_ALLOC,	/* position counter in the
						   allocated node array */
              *quadCut,				/* allocated memory */
              *quadNewCut;			/* next free entry */
   
#ifdef _MEM_STATISTICS_
   /* for the statistics */
   numCuts += numAlloc;
#endif
   
   /* increment position counter */
   position += numAlloc;
   
   /* test if new memory allocation is necessary */
   if (position >= NUM_CUT_ALLOC)
   {
      
      /* allocate memory for next NUM_CUT_ALLOC entries */
      if ((quadCut = (int *) malloc (NUM_CUT_ALLOC * sizeof (int))) ==
	  NULL)
      {
	 
	 /* memory allocation failed */
	 printf ("\n\n***** ERROR:\n\tCould not allocate space for Cuts!\n\n");
	 return (CUT_NULL);
      }
      position   = numAlloc - 1;
      quadNewCut = quadCut + numAlloc;
      
#ifdef _MEM_STATISTICS_
      /* for the statistics */
      numCutAllocs++;
#endif
   }
   else
   {
      /* increment node pointer */
      quadCut     = quadNewCut;
      quadNewCut += numAlloc;
   }
   
   return (quadCut);
}

/*** q u a d I n s i d e A l l o c
     allocates memory for a certain number (NUM_INSIDE_ALLOC, see imgetmem.h)
     of integers for entries of lines ***/

int *quadInsideAlloc (int numAlloc)
{
   static int  position = NUM_INSIDE_ALLOC,	/* position counter in the
						   allocated node array */
              *quadInside,			/* allocated memory */
              *quadNewInside;			/* next free entry */
   
#ifdef _MEM_STATISTICS_
   /* for the statistics */
   numInsides += numAlloc;
#endif
   
   /* increment position counter */
   position += numAlloc;
   
   /* test if new memory allocation is necessary */
   if (position >= NUM_INSIDE_ALLOC)
   {
      
      /* allocate memory for next NUM_INSIDE_ALLOC entries */
      if ((quadInside = (int *) malloc (NUM_INSIDE_ALLOC * sizeof (int))) ==
	  NULL)
      {
	 
	 /* memory allocation failed */
	 printf ("\n\n***** ERROR:\n\tCould not allocate space for ");
	 printf ("Insides!\n\n"); 
	 return (INSIDE_NULL);
      }
      position      = numAlloc - 1;
      quadNewInside = quadInside + numAlloc;
      
#ifdef _MEM_STATISTICS_
      /* for the statistics */
      numInsideAllocs++;
#endif
   }
   else
   {
      /* increment node pointer */
      quadInside     = quadNewInside;
      quadNewInside += numAlloc;
   }
   
   return (quadInside);
}

/*** q u a d L e a f E l e m e n t I I I A d d A l l o c
     allocates memory for a certain number (NUM_LEAF_ELEMENT_III_ADD_ALLOC,
     see imgetmem.h) of integers. ***/

int *quadLeafElementIIIAddAlloc (int numAlloc)
{
   static int  position =			/* position counter in the */
               NUM_LEAF_ELEMENT_III_ADD_ALLOC,	/* allocated node array */
              *quadLeafElementIIIAdd,		/* allocated memory */
              *quadNewLeafElementIIIAdd;	/* next free entry */
   
#ifdef _MEM_STATISTICS_
   /* for the statistics */
   numLeafElementIIIAdds += numAlloc;
#endif
   
   /* increment position counter */
   position += numAlloc;
   
   /* test if new memory allocation is necessary */
   if (position >= NUM_LEAF_ELEMENT_III_ADD_ALLOC)
   {
      
      /* allocate memory for next NUM_LEAF_ELEMENT_III_ADD_ALLOC entries */
      if ((quadLeafElementIIIAdd = (int *) malloc
	   (NUM_LEAF_ELEMENT_III_ADD_ALLOC * sizeof (int))) == NULL)
      {
	 /* memory allocation failed */
	 printf ("\n\n***** ERROR:\n\tCould not allocate space for ");
	 printf ("LeafElementIIIAdds!\n\n");
	 return (LEAF_ELEMENT_III_ADD_NULL);
      }
      position = numAlloc - 1;
      quadNewLeafElementIIIAdd = quadLeafElementIIIAdd + numAlloc;
      
#ifdef _MEM_STATISTICS_
      /* for the statistics */
      numLeafElementIIIAddAllocs++;
#endif
   }
   else
   {
      /* increment pointer */
      quadLeafElementIIIAdd     = quadNewLeafElementIIIAdd;
      quadNewLeafElementIIIAdd += numAlloc;
   }
   
   return (quadLeafElementIIIAdd);
}
