/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
/*
                          promis / imsil / src
                          --------------------
                                imquad.c

  This file contains the routines for the discretization of a given geometry
  by a finite QUADTREE. This quadtree will later be used for the geometry
  checks by the Monte Carlo Simulator for Ion Implantation in crystalline
  targets IMSIL.

  AUTHORS:   HS  Hannes Stippel, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HS  1994-02-01  new from scratch
  ===========================================================================*/

#include <stdio.h>
#include <math.h>
#include "imquad.h"
#include "imgetmem.h"
#include "imtime.h"

/* --------------------------------------------------------------------------*/

/*** q u a d G e t R e g i o n
     determines the region in which a given point lies by the intersection
     method. Usually only howManyFaces are tested; only if this equals zero,
     then all faces are tested. The number of the region where the point is
     in is returned. If the point lies outside the simulation area, zero is
     returned. ***/

int quadGetRegion (
		   long   numFaces,		/* number of faces */
		   long   faceEndIndex[],	/* end indices of the faces */
		   long   lineIndex[],		/* indices of face points */
		   double pcoord[],		/* coordinates of the points */
		   Point  testPoint		/* point in the region */
		   )
{
   int    iBegin,		/* pointer to the begin of the face */
          iFace,		/* face counter */
          iLine,		/* line counter */
          iPoint,		/* point counter */
          numIntersections;	/* number of intersections */
   Point  point1, point2;	/* points defining one edge of a face */
   double lambda, mu,		/* distances of the intersection point */
          vectorX, vectorZ;	/* vector from point1 to point2 */
   
   /* initialize the begin of the face to be investigated */
   iBegin =  0;
   
   for (iFace = 0; iFace < numFaces; iFace++)
   {
      numIntersections = 0;
      
      /* get the first point of the first line */
      iPoint = 2 * (lineIndex[iBegin] - 1);
      point1.x = pcoord[iPoint];
      point1.z = pcoord[iPoint+1];
      
      /* now test the lines of the face */
      for (iLine = iBegin+1; iLine < faceEndIndex[iFace]; iLine++)
      {
	 /* get the end point of the line */
	 iPoint = 2 * (lineIndex[iLine] - 1);
	 point2.x = pcoord[iPoint];
	 point2.z = pcoord[iPoint+1];
	 vectorX  = point2.x - point1.x;
	 vectorZ  = point2.z - point1.z;
	 
	 if (vectorX != 0.0)
	 {
	    lambda = (testPoint.x - point1.x) / vectorX;
	    /* check on possible intersection */
	    if (0.0 <= lambda && lambda < 1.0)
	    {
	       /* intersection is possible */
	       mu = point1.z - testPoint.z + lambda * vectorZ;
	       if (0.0 < mu)
		  /* this is an intersection */
		  numIntersections++;
	    }
	 }
	 
	 /* shift the starting point */
	 point1.x = point2.x;
	 point1.z = point2.z;
      }
      
      /* the last line */
      iPoint = 2 * (lineIndex[iBegin] - 1);
      point2.x = pcoord[iPoint];
      point2.z = pcoord[iPoint+1];
      vectorX  = point2.x - point1.x;
      vectorZ  = point2.z - point1.z;
      
      if (vectorX != 0.0)
      {
	 lambda = (testPoint.x - point1.x) / vectorX;
	 /* check on possible intersection */
	 if (0.0 <= lambda && lambda < 1.0)
	 {
	    /* intersection is possible */
	    mu = point1.z - testPoint.z + lambda * vectorZ;
	    if (0.0 < mu)
	       /* this is an intersection */
	       numIntersections++;
	 }
      }
      
      /* test, if point is in this face */
      if ((numIntersections % 2) == 1)
	 /* found the right face !!! */
	 return (iFace+1);
      
      /* set the begin of the next face */
      iBegin = faceEndIndex[iFace];
   }
   
   /* point is in the air region */
   return (0);
}

/*** q u a d M e a s u r e S e c t o r
     gives a strictly monotonic measure for the angle between two vectors
     (testPoint0,testPoint1) and (testPoint0,testPoint2) when proceeding
     counter clockwise around testPoint0. (Algorithm by kind permission of
     Claus Fischer) ***/

double quadMeasureSector (
			  Point testPoint0,	/* common point of the lines */
			  Point testPoint1,	/* end point of first line */
			  Point testPoint2)	/* end point of second line */
{
   double a[2], b[2], n[2], x, y, help;
   
   a[0] = testPoint1.x - testPoint0.x;
   a[1] = testPoint1.z - testPoint0.z;
   b[0] = testPoint2.x - testPoint0.x;
   b[1] = testPoint2.z - testPoint0.z;
   
   n[0] = -a[1];
   n[1] =  a[0];
   
   x = a[0]*b[0]+a[1]*b[1];
   y = n[0]*b[0]+n[1]*b[1];

   if ((x == 0.0)&&(y == 0.0)) return (0.0);
   
   help = ((fabs(x)>fabs(y)) ?
           ( y/x + ((x>0.0) ? ((y>0.0) ? 0.0 : 8.0) : 4.0)) :
           (-x/y + ((y>0.0) ? 2.0 : 6.0)));
   
   return (help);
}

/*** q u a d M a k e R e f e r e n c e s
     provides the additionally needed references for the construction of
     the quadtree. Return values: 0 ... OK ***/

int quadMakeReferences(
		       long    numFaces,	/* number of faces */
		       long    faceEndIndex[],	/* endindices for the faces
						   in the lineIndex array */
		       long    lineIndex[],	/* lines constructed by
						   references over the lines */
		       int    *numLines,	/* number of lines */
		       LineType pointIndex[],	/* lines constructed by
						   references over the
						   points */
		       int     borderLine[],	/* line is border between
						   which regions */
		       long    numPoints,	/* number of points */
		       int     pointEndIndex[],	/* endindices for the
						   pointIsInLine array */
		       int     pointIsInLine[],	/* point is in which line */
		       double  pcoord[]		/* coordinates of the point */
		       )
{
   int   iFace,		/* counter for faces */
         iBegin,	/* index to first line of the face */
         iEdge, iLine,	/* counter for lines */
         iPoint,	/* counter for points */
         iSwap,		/* help field for swapping */
         iHelp,		/* basically the same as iSwap */
         iNew, iOld,	/* old/new field for shifting according to angle */
         numResort,	/* number of lines to be shifted */
	 iOffset;	/* shifting offset */
   Point testPoint0,	/* testPoints needed for the decision, */
         testPoint1,	/* whether one angle is larger than pi */
         testPoint2;	/* than the pointIsInLine array is resorted */

   /** =================== **
       construct the lines 
    ** =================== **/
   
   /* initialize the output arrays and counter */
   for (iLine = 0; iLine < *numLines; iLine++)
   {
      pointIndex[iLine].from   = -1;
      pointIndex[iLine].to     = -1;
      borderLine[2*iLine]      = -1;
      borderLine[2*iLine+1]    = -1;
      pointIsInLine[2*iLine]   = -1;
      pointIsInLine[2*iLine+1] = -1;
   }
   *numLines = 0;
   for (iPoint = 0; iPoint < numPoints; iPoint++)
      pointEndIndex[iPoint] = 0;
   
   /* begin with the first line */
   iBegin = 1;
   
   /* loop over all faces */
   for (iFace = 0; iFace < numFaces; iFace++)
   {
      /* loop over all lines of a face */
      for (iEdge = iBegin; iEdge < faceEndIndex[iFace]; iEdge++)
      {
	 /* search in the already stored lines for the new one */
	 for (iLine = 0; iLine < *numLines; iLine++)
	    /* look, if the line is already in the array */
	    if ((pointIndex[iLine].from == lineIndex[iEdge-1] &&
		 pointIndex[iLine].to   == lineIndex[iEdge])  ||
		(pointIndex[iLine].from == lineIndex[iEdge]   &&
		 pointIndex[iLine].to   == lineIndex[iEdge-1]))
	    {
	       /* line is already in the array ==>
		  set the second regions for this line */
	       borderLine[2*iLine+1] = iFace + 1;
	       break;
	    }
	 
	 if (iLine == *numLines)
	 {
	    /* new line */
	    pointIndex[iLine].from = lineIndex[iEdge-1];
	    pointIndex[iLine].to   = lineIndex[iEdge];
	    borderLine[2*iLine]    = iFace + 1;
	    borderLine[2*iLine+1]  = 0;
	    (*numLines)++;
	 }
      }

      /* LAST LINE!!!
	 search in the already stored lines for the new one */
      for (iLine = 0; iLine < *numLines; iLine++)
	 /* look, if the line is already in the array */
	 if ((pointIndex[iLine].from == lineIndex[iEdge-1]   &&
	      pointIndex[iLine].to   == lineIndex[iBegin-1]) ||
	     (pointIndex[iLine].from == lineIndex[iBegin-1]  &&
	      pointIndex[iLine].to   == lineIndex[iEdge-1]))
	 {
	    /* line is already in the array ==>
	       set the second regions for this line */
	    borderLine[2*iLine+1] = iFace + 1;
	    break;
	 }
      
      if (iLine == *numLines)
      {
	 /* new line */
	 pointIndex[iLine].from = lineIndex[iEdge-1];
	 pointIndex[iLine].to   = lineIndex[iBegin-1];
	 borderLine[2*iLine]    = iFace + 1;
	 borderLine[2*iLine+1]  = 0;
	 (*numLines)++;
      }
      
      /* set the begin for the next face */
      iBegin = faceEndIndex[iFace] + 1;
   }
   
   /* count the number of occurences of the point in the lines */
   for (iLine = 0; iLine < *numLines; iLine++)
   {
      (pointEndIndex[pointIndex[iLine].from-1])++;
      (pointEndIndex[pointIndex[iLine].to-1])++;
   }
   /* make the end index array from the counter array */
   for (iPoint = 1; iPoint < numPoints; iPoint++)
      pointEndIndex[iPoint] += pointEndIndex[iPoint-1];
   
   /* enter the line indices */
   for (iLine = 0; iLine < *numLines; iLine++)
   {
      /* the start point of the line */
      for (iPoint = (pointIndex[iLine].from == 1) ? 0 :
	            pointEndIndex[pointIndex[iLine].from-2];
	   iPoint < pointEndIndex[pointIndex[iLine].from-1] &&
	   pointIsInLine[iPoint] != -1; iPoint++);
      pointIsInLine[iPoint] = iLine + 1;
      
      /* the end point of the line */
      for (iPoint = (pointIndex[iLine].to == 1) ? 0 :
	            pointEndIndex[pointIndex[iLine].to-2];
	   iPoint < pointEndIndex[pointIndex[iLine].to-1] &&
	   pointIsInLine[iPoint] != -1; iPoint++);
      pointIsInLine[iPoint] = iLine + 1;
   }
   
   /* test, whether the lines are sorted counter clockwise and swap, if not.
      NOTE: This only works, if the geometry is defined counter clockwise!!! */
   iLine = 0;
   for (iPoint = 0; iPoint < numPoints; iPoint++)
   {
      if (pointIndex[(pointIsInLine[iLine]-1)].from != 
	  pointIndex[(pointIsInLine[iLine+1]-1)].to)
      {
	 /* swap the lines */
	 iSwap = pointIsInLine[iLine];
	 pointIsInLine[iLine] = pointIsInLine[iLine+1];
	 pointIsInLine[iLine+1] = iSwap;
      }
      
      /* next point */
      iLine = pointEndIndex[iPoint];
   }
   
   /* now sort the lines starting at point iPoint so, that they define
      subsequent faces */
   iBegin = 0;
   for (iPoint = 0; iPoint < numPoints; iPoint++)
   {
      for (iLine = iBegin; iLine < pointEndIndex[iPoint]-1; iLine++)
      {
	 if ((borderLine[2*(pointIsInLine[iLine]-1)]     !=
	      borderLine[2*(pointIsInLine[iLine+1]-1)])     &&
	     (borderLine[2*(pointIsInLine[iLine]-1)+1]   !=
	      borderLine[2*(pointIsInLine[iLine+1]-1)+1])   &&
	     (borderLine[2*(pointIsInLine[iLine]-1)]     !=
	      borderLine[2*(pointIsInLine[iLine+1]-1)+1])   &&
	     (borderLine[2*(pointIsInLine[iLine]-1)+1]   !=
	      borderLine[2*(pointIsInLine[iLine+1]-1)]))
	 {
	    /* search for the line which is the complimentary border */
	    for (iEdge = iLine+2; iEdge < pointEndIndex[iPoint]; iEdge++)
	       if ((borderLine[2*(pointIsInLine[iLine]-1)]   ==
		    borderLine[2*(pointIsInLine[iEdge]-1)])     ||
		   (borderLine[2*(pointIsInLine[iLine]-1)+1] ==
		    borderLine[2*(pointIsInLine[iEdge]-1)+1])   ||
		   (borderLine[2*(pointIsInLine[iLine]-1)]   ==
		    borderLine[2*(pointIsInLine[iEdge]-1)+1])   ||
		   (borderLine[2*(pointIsInLine[iLine]-1)+1] ==
		    borderLine[2*(pointIsInLine[iEdge]-1)]))
		  /* this is the right line */
		  break;
	    /* swap the two entries iLine+1 and iEdge */
	    iSwap = pointIsInLine[iLine+1];
	    pointIsInLine[iLine+1] = pointIsInLine[iEdge];
	    pointIsInLine[iEdge] = iSwap;
	 }
      }
      iBegin = pointEndIndex[iPoint];
   }

   /* see, if one of the angles between subsequent lines (except the one
      between the last and the first one) is larger than pi (this leads to
      one non-convex region in the leaf of the quadtree) */
   iBegin = 0;
   for (iPoint = 0; iPoint < numPoints; iPoint++)
   {
      /* get the point to be tested */
      testPoint0.x = pcoord[2*iPoint];
      testPoint0.z = pcoord[2*iPoint+1];
      
      /* get the endpoint of the first line */
      if (pointIndex[(pointIsInLine[iBegin]-1)].from != iPoint+1)
      {
	 testPoint1.x =
	    pcoord[2*(pointIndex[(pointIsInLine[iBegin]-1)].from-1)];
	 testPoint1.z =
	    pcoord[2*(pointIndex[(pointIsInLine[iBegin]-1)].from-1)+1];
      }
      else
      {
	 testPoint1.x =
	    pcoord[2*(pointIndex[(pointIsInLine[iBegin]-1)].to-1)];
	 testPoint1.z =
	    pcoord[2*(pointIndex[(pointIsInLine[iBegin]-1)].to-1)+1];
      }
      
      /* loop over all following lines */
      iBegin++;
      for (iLine = iBegin; iLine < pointEndIndex[iPoint]; iLine++)
      {
	 /* get the endpoint of the next line, following counter clockwise */
	 if (pointIndex[(pointIsInLine[iLine]-1)].from != iPoint+1)
	 {
	    testPoint2.x =
	       pcoord[2*(pointIndex[(pointIsInLine[iLine]-1)].from-1)];
	    testPoint2.z =
	       pcoord[2*(pointIndex[(pointIsInLine[iLine]-1)].from-1)+1];
	 }
	 else
	 {
	    testPoint2.x =
	       pcoord[2*(pointIndex[(pointIsInLine[iLine]-1)].to-1)];
	    testPoint2.z =
	       pcoord[2*(pointIndex[(pointIsInLine[iLine]-1)].to-1)+1];
	 }
	 
	 /* test, if the angle between the two vectors
	    (testPoint0,testPoint1) and (testPoint0,testPoint2) is larger
	    than pi */
	 if (fabs (quadMeasureSector (testPoint0, testPoint1, testPoint2)) >
	     4.0)
	 {
	    /* resort the entries in the pointIsInLine array accordingly  -->
	       the entries must start with the line (testPoint0,testPoint2);
	       which is currently stored at iLine+1 */
	    numResort = pointEndIndex[iPoint] - (iBegin - 1);
	    iOffset   = pointEndIndex[iPoint] - iLine;
	    iSwap = pointIsInLine[iLine];
	    for (iNew = 0, iOld = iLine-(iBegin-1); iNew != iLine-(iBegin-1);)
	    {
	       iHelp = pointIsInLine[iNew+(iBegin-1)];
	       pointIsInLine[iNew+(iBegin-1)] = iSwap;
	       iSwap = iHelp;
	       iOld = iNew;
	       iNew = (iNew + iOffset) % numResort;
	    }
	    pointIsInLine[iNew+(iBegin-1)] = iSwap;
	 }
	 
	 /* set the first test point for the next test */
	 testPoint1.x = testPoint2.x;
	 testPoint1.z = testPoint2.z;
      }
      
      /* next point */
      iBegin = pointEndIndex[iPoint];
   }
   
   return (0);
}

/*** t e s t E n t i t y
     tests, whether more than one point lies within one quadtree element.
     Return values: 0             ... nothing is inside this rectangle
                    ENTITY_CUT    ... one line intersects this rectangle
		    ENTITY_INSIDE ... one point lies inside the rectangle
		    else          ... more then one line intersects the
		                      rectangle or more then one point lies
				      inside ***/

int quadTestEntity (
		    int     numLines,		/* number of lines */
		    LineType pointIndex[],	/* lines constructed by
						   references over the
						   points */
		    int    *numCuts,		/* number of intersections */
		    int     cut[],		/* intersecting lines */
		    long    numPoints,		/* number of points */
		    double  pcoord[],		/* coordinates of the point */
		    int    *numInsides,		/* number of points inside
						   the rectangle */
		    int     inside[],		/* points inside the
						   rectangle */
		    Point   center,		/* center of rectangle */
		    double  sideX,		/* sidelength x */
		    double  sideZ		/* sidelength z */
		    )
{
   int    iCut,				/* counter for intersecting lines */
          iInside,			/* counter for points inside */
          returnValueCut = ENTITY_NO,	/* return value for intersections */
          returnValueInside = ENTITY_NO,/* return value for points */
          oldNumCuts = *numCuts,	/* number of intersections */
          oldNumInsides	= *numInsides,	/* number of point inside rectangle */
          iIntersect,			/* found an intersection */
          iLine,			/* line index */
          iPointFrom, iPointTo, iPoint;	/* point indices */
   double newSideX = sideX / 2.0,	/* new sidelength */
          newSideZ = sideZ / 2.0,	/* new sidelength */
          leftBorder, rightBorder,	/* left/right boundary of rectangle */
          upperBorder, lowerBorder,	/* north/south boundary of rectangle */
          vectorX, vectorZ,		/* components of the vector of line */
	  slope,			/* slope of the line */
          testCoordinate;		/* test coordinate for cross point */
   
   /* determine the borders */
   leftBorder  = center.x - newSideX;
   rightBorder = center.x + newSideX;
   lowerBorder = center.z - newSideZ;
   upperBorder = center.z + newSideZ;
   
   /* test on intersections */
   *numCuts = 0;
   for (iCut = 0; iCut < oldNumCuts; iCut++)
   {
      /* get the line to be tested and the relevant point indices */
      iLine      = cut[iCut] - 1;
      iPointFrom = 2 * (pointIndex[iLine].from - 1);
      iPointTo   = 2 * (pointIndex[iLine].to   - 1);
      
      /* preliminary: both points must be outside the rectangle !!! */
      if ((leftBorder           > pcoord[iPointTo]     ||
	   pcoord[iPointTo]     > rightBorder          ||
	   lowerBorder          > pcoord[iPointTo+1]   ||
	   pcoord[iPointTo+1]   > upperBorder)         &&
	  (leftBorder           > pcoord[iPointFrom]   ||
	   pcoord[iPointFrom]   > rightBorder          ||
	   lowerBorder          > pcoord[iPointFrom+1] ||
	   pcoord[iPointFrom+1] > upperBorder))
      {
	 iIntersect = 0;
	 
	 /* determine the vector of the line (iPointFrom,iPointTo) */
	 vectorX = pcoord[iPointTo]   - pcoord[iPointFrom];
	 vectorZ = pcoord[iPointTo+1] - pcoord[iPointFrom+1];
	 
	 /* ===== vertical borders ===== */
	 /* left border of the rectangle */
	 if (vectorX == 0.0)
	 {
	    /* parallel line */
	    if (leftBorder       < pcoord[iPointTo]    &&
		pcoord[iPointTo] < rightBorder         &&
		((lowerBorder > pcoord[iPointTo+1]     &&
		  upperBorder < pcoord[iPointFrom+1])  ||
		 (lowerBorder > pcoord[iPointFrom+1]   &&
		  upperBorder < pcoord[iPointTo+1])))
	    {
	       /* line intersects the rectangle */
	       cut[*numCuts] = cut[iCut];
	       (*numCuts)++;
	       returnValueCut += ENTITY_CUT;
	       iIntersect = 1;
	    }
	 }
	 else	/* vectorX != 0.0 */
	 {
	    /* compute the test coordinate for the left border */
	    slope = (leftBorder - pcoord[iPointFrom]) / vectorX;
	    if (0.0 < slope && slope < 1.0)
	    {
	       /* intersection is possible */
	       testCoordinate = pcoord[iPointFrom+1] + slope * vectorZ;
	       if (lowerBorder < testCoordinate &&
		   testCoordinate < upperBorder)
	       {
		  /* line intersects the rectangle */
		  cut[*numCuts] = cut[iCut];
		  (*numCuts)++;
		  returnValueCut += ENTITY_CUT;
		  iIntersect = 1;
	       }
	    }
	    
	    if (iIntersect == 0)
	    {
	       /* compute the test coordinate for the right border */
	       slope = (rightBorder - pcoord[iPointFrom]) / vectorX;
	       if (0.0 < slope && slope < 1.0)
	       {
		  /* intersection is possible */
		  testCoordinate = pcoord[iPointFrom+1] + slope * vectorZ;
		  if (lowerBorder < testCoordinate &&
		      testCoordinate < upperBorder)
		  {
		     /* line intersects the rectangle */
		     cut[*numCuts] = cut[iCut];
		     (*numCuts)++;
		     returnValueCut += ENTITY_CUT;
		     iIntersect = 1;
		  }
	       }
	    }
	 }
	 
	 if (iIntersect == 0)
	 {
	    /* ===== horizontal borders ===== */
	    if (vectorZ == 0.0)
	    {
	       /* parallel line */
	       if (lowerBorder        < pcoord[iPointTo+1] &&
		   pcoord[iPointTo+1] < upperBorder        &&
		   ((leftBorder  > pcoord[iPointTo]        &&
		     rightBorder < pcoord[iPointFrom])     ||
		    (leftBorder  > pcoord[iPointFrom]      &&
		     rightBorder < pcoord[iPointTo])))
	       {
		  /* line intersects the rectangle */
		  cut[*numCuts] = cut[iCut];
		  (*numCuts)++;
		  returnValueCut += ENTITY_CUT;
	       }
	    }
	    else	/* vectorZ != 0.0 */
	    {
	       /* compute the test coordinate for the lower border */
	       slope = (lowerBorder - pcoord[iPointFrom+1]) / vectorZ;
	       if (0.0 < slope && slope < 1.0)
	       {
		  /* intersection is possible */
		  testCoordinate = pcoord[iPointFrom] + slope * vectorX;
		  if (leftBorder     < testCoordinate &&
		      testCoordinate < rightBorder)
		  {
		     /* line intersects the rectangle */
		     cut[*numCuts] = cut[iCut];
		     (*numCuts)++;
		     returnValueCut += ENTITY_CUT;
		     iIntersect = 1;
		  }
	       }
	       
	       if (iIntersect == 0)
	       {
		  /* compute the test coordinate for the upper border */
		  slope = (upperBorder - pcoord[iPointFrom+1]) / vectorZ;
		  if (0.0 < slope && slope < 1.0)
		  {
		     /* intersection is possible */
		     testCoordinate = pcoord[iPointFrom] + slope * vectorX;
		     if (leftBorder     < testCoordinate &&
			 testCoordinate < rightBorder)
		     {
			/* line intersects the rectangle */
			cut[*numCuts] = cut[iCut];
			(*numCuts)++;
			returnValueCut += ENTITY_CUT;
		     }
		  }
	       }
	    }
	 }		/* no intersection with vertical borders */
      }		/* both points are outside the rectangle */
      else
      {
	 if ((leftBorder           <= pcoord[iPointTo]     &&
	      pcoord[iPointTo]     <= rightBorder          &&
	      lowerBorder          <= pcoord[iPointTo+1]   &&
	      pcoord[iPointTo+1]   <= upperBorder)         ||
	     (leftBorder           <= pcoord[iPointFrom]   &&
	      pcoord[iPointFrom]   <= rightBorder          &&
	      lowerBorder          <= pcoord[iPointFrom+1] &&
	      pcoord[iPointFrom+1] <= upperBorder))
	 {
	    /* at least one point is inside the rectangle ==> this line has
	       still to be considered for the next intersections!!! */
	    cut[*numCuts] = cut[iCut];
	    (*numCuts)++;
	 }
      }
   }	/* next line */
   
   /* this is a ?*dirty*! trick to enable easier testing afterwards */
   if (returnValueCut == 2 * ENTITY_CUT)
      returnValueCut += ENTITY_NO_OK;
   
   /* test, whether a point is in the rectangle */
   *numInsides = 0;
   for (iInside = 0; iInside < oldNumInsides; iInside++)
   {
      /* get the point to be tested */
      iPoint = 2 * (inside[iInside]-1);
      
      /* test the point */
      if (leftBorder       <= pcoord[iPoint]   &&
	  pcoord[iPoint]   <= rightBorder      &&
	  lowerBorder      <= pcoord[iPoint+1] &&
	  pcoord[iPoint+1] <= upperBorder)
      {
	 /* point is inside the rectangle */
	 inside[*numInsides] = inside[iInside];
	 (*numInsides)++;
	 returnValueInside += ENTITY_INSIDE;
      }
   }
   
   /* this is a ?*dirty*! trick to enable easier testing afterwards */
   if (returnValueInside == 2 * ENTITY_INSIDE)
      returnValueInside += ENTITY_NO_OK;
   
   return (returnValueCut+returnValueInside);
}

/*** q u a d M a k e T r e e
     discretizes the given geometry using a finite quadtree.
     RETURN values: NODE_NULL ... an error occured (probably not enough memory)
                    else      ... pointer to a new generated node of the
		                  tree ***/

QuadNode *quadMakeTree (
			long   numFaces,	/* number of faces */
			long   faceEndIndex[],	/* endindices for the faces
						   in the lineIndex array */
			long   lineIndex[],	/* lines constructed by
						   references over the lines */
			int    numLines,	/* number of lines */
			LineType pointIndex[],	/* lines constructed by
						   references over the
						   points */
			int    borderLine[],	/* line is border between
						   which regions */
			int    numCuts,		/* number of intersecting
						   lines */
			int    cut[],		/* intersecting lines */
			long   numPoints,	/* number of points */
			int    pointEndIndex[],	/* endindices for the
						   pointIsInLine array */
			int    pointIsInLine[],	/* point is in which line */
			double pcoord[],	/* coordinates of the point */
			int    numInsides,	/* number of points inside
						   the rectangle */
			int    inside[],	/* points inside the
						   rectangle */
			Point  center,		/* center of the rectangle */
			double sideX,		/* side x of the rectangle */
			double sideZ,		/* side z of the rectangle */
			double quadResolutionX,	/* quadtree resolution */
			double quadResolutionZ)	/* quadtree resolution */
{
   QuadNode           *quadNode;		/* new node */
   QuadLeaf           *quadLeaf;		/* new leaf */
   QuadLeafElementI   *quadLeafElementI;	/* new leaf element type I */
   QuadLeafElementII  *quadLeafElementII;	/* new leaf element type II */
   QuadLeafElementIII *quadLeafElementIII;	/* new leaf element type III */
   Point               newCenter;		/* new center point */
   double              newSideX,		/* new sidelength, x */
                       newSideZ;		/* new sidelength, z */
   int                 iChild,			/* i-th child of the node */
                       newNumCuts,		/* # of intersecting lines */
                      *newCut,			/* intersecting lines */
                       newNumInsides,		/* number of points inside */
                      *newInside,		/* points inside */
                       iCut,			/* counter for intersections */
                       iInside,			/* counter for inside points */
                       iBegin,			/* pointer to array begin */
                       iLine,			/* pointer to a line */
                       iEdge1,			/* pointer to first line */
                       iEdge2,			/* pointer to second line */
                       testEntity;		/* decide between leaf/node */ 
   
   /* provide space for new node */
   if ((quadNode = quadNodeAlloc ()) == NODE_NULL)
      /* memory allocation failed */
      return (NODE_NULL);
   
   if (numCuts > 0)
      /* provide space for new cut information */
      if ((newCut = quadCutAlloc (numCuts)) == CUT_NULL)
	 /* memory allocation failed */
	 return (NODE_NULL);
   
   if (numInsides > 0)
      /* provide space for new information about point inside the rectangle */
      if ((newInside = quadInsideAlloc (numInsides)) == INSIDE_NULL)
	 /* memory allocation failed */
	 return (NODE_NULL);
   
   /* initialize the node */
   quadNode->quadType = 0;
   
   /* determine new side length's */
   newSideX = sideX / 2.0;
   newSideZ = sideZ / 2.0;
   
   /* loop over all subnodes of the tree */
   for (iChild = 0; iChild < NUM_QUAD_CHILDREN; iChild++)
   {
      /* determine new center */
      newCenter.x = NEW_X (iChild, center.x, newSideX/2.0);
      newCenter.z = NEW_Z (iChild, center.z, newSideZ/2.0);
      
      /* initialize the intersection array */
      for (iCut = 0; iCut < numCuts; iCut ++)
	 newCut[iCut] = cut[iCut];
      newNumCuts = numCuts;
      
      /* initialize the array for the points inside the rectangle */
      for (iInside = 0; iInside < numInsides; iInside ++)
	 newInside[iInside] = inside[iInside];
      newNumInsides = numInsides;
      
      /* test on intersections */
      if ((testEntity = quadTestEntity (numLines, pointIndex, &newNumCuts,
					newCut, numPoints, pcoord,
					&newNumInsides, newInside, newCenter,
					newSideX, newSideZ)) >= ENTITY_NO_OK)
      {
	 if (newSideX > quadResolutionX || newSideZ > quadResolutionZ)
	 {
	    /* split this rectangle again */
	    SET_NODE (quadNode, iChild) =
	       quadMakeTree (numFaces, faceEndIndex, lineIndex, numLines,
			     pointIndex, borderLine, newNumCuts, newCut,
			     numPoints, pointEndIndex, pointIsInLine, pcoord,
			     newNumInsides, newInside, newCenter, newSideX,
			     newSideZ, quadResolutionX, quadResolutionZ);
	    if (quadNode == NODE_NULL)
	       return (NODE_NULL);
	    (SET_NODE(quadNode, iChild))->quadParent = quadNode;
	 }
	 else
	 {
	    /* maximum allowed accuracy reached ==> LEAF */
	    printf ("\n\n***** Warning: \n\t");
	    printf ("Minimum resolution reached ==> Can not guarantee");
	    printf ("exact geometry representation!\n\n");

	    /* allocate space for the leaf */
	    if ((quadLeaf = quadLeafAlloc ()) == LEAF_NULL)
	       return (NODE_NULL);
	    /* set the leaf */
	    SET_LEAF (quadNode, iChild) = quadLeaf;
	    /* set the type identifier for a leaf */
	    quadNode->quadType = quadNode->quadType | (1 << iChild);
	    
	    /* === set the region --> assume element type I === */
	    /* allocate space for the leaf element */
	    if ((quadLeafElementI = quadLeafElementIAlloc ()) ==
		LEAF_ELEMENT_I_NULL)
	       return (NODE_NULL);
	    /* set the leaf element */
	    SET_LEAF_ELEMENT_I (quadNode, iChild) = quadLeafElementI;
	    /* set the leaf type */
	    SET_LEAF_TYPE (quadNode, iChild) = 1;
	    /* determine the region of the leaf */
	    quadLeafElementI->region =
	       quadGetRegion (numFaces, faceEndIndex, lineIndex, pcoord,
			      newCenter);
	 }
      }
      else
      {
	 /* === no more splitting necessary === */
	 /* allocate space for the leaf */
	 if ((quadLeaf = quadLeafAlloc ()) == LEAF_NULL)
	    return (NODE_NULL);
	 /* set the leaf */
	 SET_LEAF (quadNode, iChild) = quadLeaf;
	 /* set the type identifier for a leaf */
	 quadNode->quadType = quadNode->quadType | (1 << iChild);
	 
	 /* set the region */
	 switch (testEntity)
	 {
	    /* leaf type I (only one region) */
	    case ENTITY_NO:
	       /* allocate space for the leaf element */
	       if ((quadLeafElementI = quadLeafElementIAlloc ()) ==
		   LEAF_ELEMENT_I_NULL)
		  return (NODE_NULL);
	       /* set the leaf element */
	       SET_LEAF_ELEMENT_I (quadNode, iChild) = quadLeafElementI;
	       /* set the leaf type */
	       SET_LEAF_TYPE (quadNode, iChild) = 1;
	       /* determine the region of the leaf */
	       quadLeafElementI->region =
		  quadGetRegion (numFaces, faceEndIndex, lineIndex, pcoord,
				 newCenter);
	       break;
	    
	    /* leaf type II (one line intersects the leaf ==> two regions) */
	    case ENTITY_CUT:
	       /* allocate space for the leaf element */
	       if ((quadLeafElementII = quadLeafElementIIAlloc ()) ==
		   LEAF_ELEMENT_II_NULL)
		  return (NODE_NULL);
	       /* set the leaf element */
	       SET_LEAF_ELEMENT_II (quadNode, iChild) = quadLeafElementII;
	       /* set the leaf type */
	       SET_LEAF_TYPE (quadNode, iChild) = 2;
	       /* set the regions on the left and right of the line */
	       quadLeafElementII->leftRegion  = borderLine[2*(newCut[0])-1];
	       quadLeafElementII->rightRegion = borderLine[2*(newCut[0]-1)];
	       /* set the pointer to the begin and end point of the line */
	       quadLeafElementII->fromPoint = pointIndex[newCut[0]-1].from;
	       quadLeafElementII->toPoint   = pointIndex[newCut[0]-1].to;
	       break;
	    
	    /* leaf type III one point is in the leaf */
	    case ENTITY_INSIDE:
	       /* allocate space for the leaf element */
	       if ((quadLeafElementIII = quadLeafElementIIIAlloc ()) ==
		   LEAF_ELEMENT_III_NULL)
		  return (NODE_NULL);
	       /* set the leaf element */
	       SET_LEAF_ELEMENT_III (quadNode, iChild) = quadLeafElementIII;
	       /* set the leaf type */
	       SET_LEAF_TYPE (quadNode, iChild) = 3;
	       /* set the inner point of the leaf */
	       quadLeafElementIII->innerPoint = newInside[0];
	       /* set the pointer to the begin of the array containing the
		  lines emerging from the inner point */
	       iBegin = (newInside[0] == 1) ?
		  0 : pointEndIndex[newInside[0]-2];
	       /* set the number of regions of the leaf */
	       quadLeafElementIII->numRegions =
		  pointEndIndex[newInside[0]-1] - iBegin;
	       /* allocate space for the regions */
               if ((quadLeafElementIII->region = (int *)
		    quadLeafElementIIIAddAlloc
		    (quadLeafElementIII->numRegions)) ==
		   LEAF_ELEMENT_III_ADD_NULL)
		  return (NODE_NULL);
	       /* allocate space for the indices of the end points */
               if ((quadLeafElementIII->endPoint = (int *)
		    quadLeafElementIIIAddAlloc
		    (quadLeafElementIII->numRegions)) == 
		   LEAF_ELEMENT_III_ADD_NULL)
		  return (NODE_NULL);
	       /* set the pointer to the first line of the region */
	       iEdge1 = pointIsInLine[iBegin] - 1;
	       if (quadLeafElementIII->numRegions > 2)
	       {
		  /* this is only if more than one line are emerging from
		     this point */
		  for (iLine = iBegin+1;
		       iLine < pointEndIndex[newInside[0]-1];
		       iLine++)
		  {
		     /* set the pointer to the second line of the region */
		     iEdge2 = pointIsInLine[iLine] - 1;
		     /* now, set the appropriate region */
		     if (borderLine[2*iEdge1] == borderLine[2*iEdge2])
			quadLeafElementIII->region[iLine-iBegin-1] =
			   borderLine[2*iEdge1];
		     else
			if (borderLine[2*iEdge1+1] == borderLine[2*iEdge2])
			   quadLeafElementIII->region[iLine-iBegin-1] =
			      borderLine[2*iEdge1+1];
			else
			   if (borderLine[2*iEdge1+1] ==
			       borderLine[2*iEdge2+1])
			      quadLeafElementIII->region[iLine-iBegin-1] =
				 borderLine[2*iEdge1+1];
			   else
			      if (borderLine[2*iEdge1] ==
				  borderLine[2*iEdge2+1])
				 quadLeafElementIII->region[iLine-iBegin-1] =
				    borderLine[2*iEdge1];
			      else
			      {
				 /* this should never happen */
				 quadLeafElementIII->region[iLine-iBegin-1] =
				    -1;
				 printf ("\n\n***** ERROR:\n\t");
				 printf ("Strange things happened ");
				 printf ("while searching the region!\n\n");
				 return (NODE_NULL);	 
			      }
		     
		     /* set the end point of the first line of the region */
		     if (pointIndex[iEdge1].from == newInside[0])
			quadLeafElementIII->endPoint[iLine-iBegin-1] =
			   pointIndex[iEdge1].to;
		     else
			if (pointIndex[iEdge1].to == newInside[0])
			   quadLeafElementIII->endPoint[iLine-iBegin-1] =
			      pointIndex[iEdge1].from;
			else
			{
			   /* again this should never happen */
			   quadLeafElementIII->endPoint[iLine-iBegin-1] =
			      -1;
			   printf ("\n\n***** ERROR:\n\t");
			   printf ("Strange things happened ");
			   printf ("while searching the end point!\n\n");
			   return (NODE_NULL);
			}
		     
		     /* set the next line */
		     iEdge1 = iEdge2;
		  }
		  /* set the last region */
		  iEdge2 = pointIsInLine[iBegin] - 1;
		  if (borderLine[2*iEdge1] == borderLine[2*iEdge2])
		     quadLeafElementIII->region[iLine-iBegin-1] =
			borderLine[2*iEdge1];
		  else
		     if (borderLine[2*iEdge1+1] == borderLine[2*iEdge2])
			quadLeafElementIII->region[iLine-iBegin-1] =
			   borderLine[2*iEdge1+1];
		     else
			if (borderLine[2*iEdge1+1] == borderLine[2*iEdge2+1])
			   quadLeafElementIII->region[iLine-iBegin-1] =
			      borderLine[2*iEdge1+1];
			else
			   if (borderLine[2*iEdge1] == borderLine[2*iEdge2+1])
			      quadLeafElementIII->region[iLine-iBegin-1] =
				 borderLine[2*iEdge1];
			   else
			   {
			      /* this should never happen */
			      quadLeafElementIII->region[iLine-iBegin-1] =
				 -1;
			      printf ("\n\n***** ERROR:\n\t");
			      printf ("Strange things happened ");
			      printf ("while searching the last region!\n\n");
			      return (NODE_NULL);
			   }
		  
		  /* set the last end point */
		  if (pointIndex[iEdge1].from == newInside[0])
		     quadLeafElementIII->endPoint[iLine-iBegin-1] =
			pointIndex[iEdge1].to;
		  else
		     if (pointIndex[iEdge1].to == newInside[0])
			quadLeafElementIII->endPoint[iLine-iBegin-1] =
			   pointIndex[iEdge1].from;
		     else
		     {
			/* again this should never happen */
			quadLeafElementIII->endPoint[iLine-iBegin-1] =
			   -1;
			printf ("\n\n***** ERROR:\n\t");
			printf ("Strange things happened ");
			printf ("while searching the last end point!\n\n");
			return (NODE_NULL);
		     }
	       }
	       else
	       {
		  /* only when we have two lines */
		  if (pointIndex[iEdge1].from == newInside[0])
		  {
		     quadLeafElementIII->region[0] = borderLine[2*iEdge1];
		     quadLeafElementIII->region[1] = borderLine[2*iEdge1+1];
		     quadLeafElementIII->endPoint[0] = pointIndex[iEdge1].to;
		  }
		  else
		     if (pointIndex[iEdge1].to == newInside[0])
		     {
			quadLeafElementIII->region[0] = borderLine[2*iEdge1+1];
			quadLeafElementIII->region[1] = borderLine[2*iEdge1];
			quadLeafElementIII->endPoint[0] =
			   pointIndex[iEdge1].from;
		     }
		     else
		     {
			/* this should never happen */
			quadLeafElementIII->region[0] = -1;
			quadLeafElementIII->region[1] = -1;
			printf ("\n\n***** ERROR:\n\t");
			printf ("Strange things happened while searching ");
			printf ("the first region of two lines!\n\n");
			return (NODE_NULL);
		     }
		  
		  /* the second line */
		  iEdge2 = pointIsInLine[iBegin+1] - 1;
		  if (pointIndex[iEdge2].from == newInside[0])
		     quadLeafElementIII->endPoint[1] =
			pointIndex[iEdge2].to;
		  else
		     if (pointIndex[iEdge2].to == newInside[0])
			quadLeafElementIII->endPoint[1] =
			   pointIndex[iEdge2].from;
		     else
		     {
			/* this should never happen */
			quadLeafElementIII->endPoint[1] = -1;
			printf ("\n\n***** WARNING:\n\t");
			printf ("Strange things happened while searching ");
			printf ("the last end point for two lines!\n\n");
			return (NODE_NULL);
		     }
	       }
	       break;
	    
	    default:
	       printf ("\n\n***** ERROR:\n\t");
	       printf ("Unknown result <%d> of function quadTestEntity!\n\n",
		       testEntity);
	    return (NODE_NULL);
	       break;
	 }
      }
   }
   
   /* return the new node */
   return (quadNode);
}

/*** q u a d M a k e A m b i e n t
     determines the closed polygon which is the border to the ambient or the
     end of the simulation area, respectively. Only the point indices are
     stored succesively defining the hull counterclockwise, to allow a fast
     acces.
     Return values: 0 ... everything is OK
                    1 ... less than three border lines
		    2 ... too many border lines found (>numAmbients) ***/

int quadMakeAmbient (
		     int  numLines,		/* number of lines */
		     LineType pointIndex[],		/* lines */
		     int  borderLine[],		/* border for which regions */
		     int  numAmbients,		/* # of lines of the hull */
		     int  ambientBorder[])	/* lines of the hull */
{
   int iLine, iAmbient,		/* counter */
       iStore,			/* last line, where point was inserted */
       iSwap,			/* used for swapping */
       numAmbientFound = 0;	/* counter for the lines of the hull */
   
   if (numAmbients < 3)
   {
      /* a polygon must have at least three lines! */
      printf ("\n\n***** ERROR:\n\t");
      printf ("Less then 3 border lines (%d) --> that's impossible!\n\n",
	      numAmbients);
      return (-1);
   }
   
   /* ============================
      determine the closed polygon
      ============================ */

   for (iLine = 0; iLine < numLines; iLine++)
      if (borderLine[2*iLine] == 0 || borderLine[2*iLine+1] == 0)
      {
	 /* found a new piece of the border */
	 if (numAmbientFound >= numAmbients)
	 {
	    /* too many ambient lines */
	    printf ("\n\n***** ERROR:\n\t");
	    printf ("Found more then %d border lines!\n\n",
		    numAmbients);
	    return (-2);
	 }
	 /* insert the new piece */
	 ambientBorder[numAmbientFound] = iLine + 1;
	 numAmbientFound++;
      }
   
   /* =================================================================
      get the pointindices of the lines and sort the hull so that it is
      closed and from the view of the surrounding ambient area
      counterclockwise.
      ================================================================= */
   
   /* fill in the first two points which come from the first line and store
      the line of ambientBorder[1] for later!
      
      Note: As the geometry is oriented counterclockwise, we must insert the
      reverse line in order to obtain the right orientation for the ambient
      area! */
   iAmbient = ambientBorder[0];
   ambientBorder[0] = pointIndex[iAmbient-1].to;
   iStore = ambientBorder[1];
   ambientBorder[1] = pointIndex[iAmbient-1].from;
   
   /* Now do the same with the rest of the border lines */
   for (iAmbient = 2; iAmbient < numAmbientFound; iAmbient++)
   {
      /* first see, if the stored line contains the requested point */
      if (pointIndex[iStore-1].to == ambientBorder[iAmbient-1])
      {
	 /* this is the next point --> save the currently in
	    ambientBorder stored line information and insert the detected
	    point into the array */
	 iSwap  = iStore;
	 iStore = ambientBorder[iAmbient];
	 ambientBorder[iAmbient] = pointIndex[iSwap-1].from;
      }
      else
	 /* this part should probably only be used when the geometry is
	    oriented clockwise??? -- So usually not!!! */
	 if (pointIndex[iStore-1].from == ambientBorder[iAmbient-1])
	 {
	    /* this is the next point --> save the currently in
	       ambientBorder stored line information and insert the detected
	       point into the array */
	    iSwap  = iStore;
	    iStore = ambientBorder[iAmbient];
	    ambientBorder[iAmbient] = pointIndex[iSwap-1].to;
	 }
	 else
	 {
	    /* now we have to look in the other lines, if we can find the
	       adjacent line */
	    for (iLine = iAmbient; iLine < numAmbientFound; iLine++)
	    {
	       if (pointIndex[ambientBorder[iLine]-1].to ==
		   ambientBorder[iAmbient-1])
	       {
		  /* this is the next point --> move the currently stored
		     line and insert the detected point into the array at
		     this index. */
		  iSwap = ambientBorder[iLine] - 1;
		  if (iLine != iAmbient)
		     ambientBorder[iLine] = ambientBorder[iAmbient];
		  ambientBorder[iAmbient] = pointIndex[iSwap].from;
		  break;
	       }
	       else
		  if (pointIndex[ambientBorder[iLine]-1].from ==
		   ambientBorder[iAmbient-1])
		  {
		     /* this is the next point --> move the currently stored
			line and insert the detected point into the array at
			this index.
			Note: Again, this shoul NEVER happen; only for wrong
			oriented structures. */
		     iSwap = ambientBorder[iLine] - 1;
		     if (iLine != iAmbient)
			ambientBorder[iLine] = ambientBorder[iAmbient];
		     ambientBorder[iAmbient] = pointIndex[iSwap].to;
		     break;
		  }
	    }
	    if (iLine == numAmbientFound)
	    {
	       /* error: no adjecent line found?!? */
	       printf ("\n\n***** ERROR:\n\t");
	       printf ("No adjecent line found for point <%d>!\n\n",
		       ambientBorder[iAmbient-1]);
	       return (-3);
	    }
	 }
   }
   
   /* everything OK */
   return (0);
}
