/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
/*
                          promis / imsil / src
                          --------------------
                                imquad.h

  Header file for the QUADTREE generation.

  AUTHORS:   HS  Hannes Stippel, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HS  1994-02-01  new from scratch
  ===========================================================================*/

/* --------------------------------------------------------------------------*/

/* constants 
   ========= */

#define ENTITY_NO		0
#define ENTITY_CUT		1
#define ENTITY_INSIDE		2
#define ENTITY_NO_OK		ENTITY_CUT+ENTITY_INSIDE

#define TYPE_I			1
#define TYPE_II			2
#define TYPE_III		3
#define NUM_QUAD_CHILDREN	4

/* macros 
   ====== */

#define SET_NODE(n,i)			((n)->quadChild[(i)].quadNode)
#define SET_LEAF(l,i)			((l)->quadChild[(i)].quadLeaf)
#define SET_LEAF_ELEMENT_I(l,i)		(((l)->quadChild[(i)].quadLeaf)->quadLeafElement.quadLeafElementI)
#define SET_LEAF_ELEMENT_II(l,i)	(((l)->quadChild[(i)].quadLeaf)->quadLeafElement.quadLeafElementII)
#define SET_LEAF_ELEMENT_III(l,i)	(((l)->quadChild[(i)].quadLeaf)->quadLeafElement.quadLeafElementIII)
#define SET_LEAF_TYPE(l,i)		(((l)->quadChild[(i)].quadLeaf)->quadLeafType)
#define IS_LEAF(n,i)			(((n)->quadType) & (1 << (i)))
#define GET_NODE(n,i)			(IS_LEAF ((n), (i)) ? NODE_NULL : \
					 (n)->quadChild[(i)].quadNode)
#define GET_LEAF(l,i)			(IS_LEAF ((l), (i)) ? \
					 (l)->quadChild[(i)].quadLeaf : \
					 NODE_NULL)
#define NEW_X(i,c,s)			(((i) < 2) ? ((c) - (s)) : ((c) + (s)))
#define NEW_Z(i,c,s)			(((i) == 0 || (i) == 2) ? \
					 ((c) - (s)) : ((c) + (s)))

/* --------------------------------------------------------------------------*/

/* external global variables
   ========================= */

#ifdef _MEM_STATISTICS_
/* === These are the global counter for the memory statistics === */
extern int numNodes;			/* # of allocated nodes */
extern int numNodeAllocs;		/* # of allocations for the nodes */
extern int numLeafs;			/* # of allocated leafs */
extern int numLeafAllocs;		/* # of allocations for the leafs */
extern int numLeafElementIs;		/* # of allocated elements Type I */
extern int numLeafElementIAllocs;	/* # of allocs for the element I */
extern int numLeafElementIIs;		/* # of allocated elements Type II */
extern int numLeafElementIIAllocs;	/* # of allocs for the element II */
extern int numLeafElementIIIs;		/* # of allocated elements Type III */
extern int numLeafElementIIIAllocs;	/* # of allocs for the element III */
extern int numCuts;			/* # of allocated cutinformations */
extern int numCutAllocs;		/* # of allocations for cuts */
extern int numInsides;			/* # of allocated insides */
extern int numInsideAllocs;		/* # of allocations for inside */
extern int numLeafElementIIIAddAllocs;	/* # of add. allocs for element III */
extern int numLeafElementIIIAdds;	/* # of add. allocated elements III */
#endif

/* --------------------------------------------------------------------------*/

/* global variables */
double minX,maxX,minZ,maxZ;

/* structures and type declarations 
   ================================ */

typedef struct				/* the points */
{
   double x, z;				/* x/z coordinate of the point */
}
Point;

typedef struct				/* the lines */
{
   int from, to;			/* references to the points:
					      from = begin of line
					      to   = end   of line */
}
LineType;

typedef struct QuadLeafElementStructI	/* Type I:   only one region */
{
   int region;				/* the region index */
}
QuadLeafElementI;

typedef struct QuadLeafElementStructII	/* Type II:  two regions ==> one line
					             intersects the leaf */
{
   int leftRegion, rightRegion,		/* regions on the left/right side of
					   the line */
       fromPoint,  toPoint;		/* line goes from Point with index
					   fromPoint to toPoint */
}
QuadLeafElementII;

typedef struct QuadLeafElementStructIII	/* Type III: n regions ==> point is
					             inside the leaf */
{
   int  numRegions,			/* number of regions in the leaf */
       *region,				/* the regions in the leaf */
        innerPoint,			/* the point inside the leaf */
       *endPoint;			/* array of endpoints of the lines
					   starting from innerPoint */
}
QuadLeafElementIII;

typedef struct QuadLeafStruct		/* one leaf of the quadtree */
{
   int  quadLeafType;			/* type of the leaf */
   union
   {
      struct QuadLeafElementStructI   *quadLeafElementI;	/* Type I */
      struct QuadLeafElementStructII  *quadLeafElementII;	/* Type II */
      struct QuadLeafElementStructIII *quadLeafElementIII;	/* Type III */
   }
   quadLeafElement;			/* pointer to the element */
}
QuadLeaf;

typedef struct QuadNodeStruct 		/* one node of the quadtree */
{
   int quadType;			/* type of the nodes */
   union
   {
      struct QuadNodeStruct *quadNode;	/* node of the quadtree */
      struct QuadLeafStruct *quadLeaf;	/* leaf of the quadree */
   }
   quadChild[NUM_QUAD_CHILDREN];	/* children (order is x before z) */
   struct QuadNodeStruct *quadParent;	/* parent node */
}
QuadNode;

/* --------------------------------------------------------------------------*/

/* useful NULL pointer
   =================== */

#define NODE_NULL			(QuadNode *) NULL
#define LEAF_NULL			(QuadLeaf *) NULL
#define LEAF_ELEMENT_I_NULL		(QuadLeafElementI *) NULL
#define LEAF_ELEMENT_II_NULL		(QuadLeafElementII *) NULL
#define LEAF_ELEMENT_III_NULL		(QuadLeafElementIII *) NULL
#define LEAF_ELEMENT_III_ADD_NULL	(int *) NULL
#define CUT_NULL			(int *) NULL
#define INSIDE_NULL			(int *) NULL

/* --------------------------------------------------------------------------*/

/* function prototypes 
   =================== */

/* this is the function that generates the quad tree */
QuadNode *quadMakeTree (
			long   numFaces,	/* number of faces */
			long   faceEndIndex[],	/* endindices for the faces
						   in the lineIndex array */
			long   lineIndex[],	/* lines constructed by
						   references over the lines */
			int    numLines,	/* number of lines */
			LineType pointIndex[],	/* lines constructed by
						   references over the
						   points */
			int    borderLine[],	/* line is border between
						   which regions */
			int    numCuts,		/* number of intersecting
						   lines */
			int    cut[],		/* intersecting lines */
			long   numPoints,	/* number of points */
			int    pointEndIndex[],	/* endindices for the
						   pointIsInLine array */
			int    pointIsInLine[],	/* point is in which line */
			double pcoord[],	/* coordinates of the point */
			int    numInsides,	/* number of points inside
						   the rectangle */
			int    inside[],	/* points inside the
						   rectangle */
			Point  center,		/* center of the rectangle */
			double sideX,		/* side x of the rectangle */
			double sideZ,		/* side z of the rectangle */
			double quadResolutionX,	/* quadtree resolution */
			double quadResolutionZ);/* quadtree resolution */

/* generate the hull of the simulation area */
int quadMakeAmbient (
		     int  numLines,		/* number of lines */
		     LineType pointIndex[],		/* lines */
		     int  borderLine[],		/* border for which regions */
		     int  numAmbients,		/* # of lines of the hull */
		     int  ambientBorder[]);	/* lines of the hull */


/*** q u a d M a k e R e f e r e n c e s
     provides the additionally needed references for the construction of
     the quadtree. Return values: 0 ... OK ***/
int quadMakeReferences(
                       long    numFaces,        /* number of faces */
                       long    faceEndIndex[],  /* endindices for the faces
                                                   in the lineIndex array */
                       long    lineIndex[],     /* lines constructed by
                                                   references over the lines */
                       int    *numLines,        /* number of lines */
                       LineType pointIndex[],   /* lines constructed by
                                                   references over the
                                                   points */
                       int     borderLine[],    /* line is border between
                                                   which regions */
                       long    numPoints,       /* number of points */
                       int     pointEndIndex[], /* endindices for the
                                                   pointIsInLine array */
                       int     pointIsInLine[], /* point is in which line */
                       double  pcoord[]         /* coordinates of the point */
                       );
