/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
/*
                              S E S A M E
                          --------------------
                              imsurfgeo.c

  Routines for geometric handling in the near surface region.


  AUTHORS:   HW  Horst Wagner, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HW  1994-10-14  new from scratch
  ===========================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "imquad.h"
#include "imsurfgeo.h"

/* global variables */

static double *SurfCoordX,*SurfCoordZ,*SurfVecX,*SurfVecZ,*SurfLen,*SurfSumLen;
static int numSurfaceBorders;

/*** i m S u r f M a k e

      initializes and allocates structures for use of imsurf routines

***/ 	


int imsurfMake (
		   int        numAmbients,	/* # of lines for the hull */
		   int        ambientBorder[],	/* lines of simulation hull */
		   double     pcoord[],		/* coordinates of the points */
      		   double     minX,double maxX, /* Maximum coordinates of- */
                   double     minZ,double maxZ) /* Simulation Area */
{
   int    iPoint,		/* indices of points defining the line */
    	  iLeftSurf,iRightSurf, /* indice of left and right edge of surface */       
	  iAmbient;		/* counter */

   double dX,dZ,
	  minXminZ,maxXminZ;	


   /* Extract start (right) and end (left) of the surface */	

   maxXminZ = maxZ;
   minXminZ = maxZ;
#ifdef _DEBUG_SURFGEO_
      printf("Debug Imsurfgeo.c (_DEBUG_SURFGEO_ defined)");       
      printf("\n\nCoordinates of Ambient points:\n");
#endif
      
   for (iAmbient = 0; iAmbient < numAmbients; iAmbient++)
   {
      /* get point of the line */
      iPoint = 2 * (ambientBorder[iAmbient] - 1);
#ifdef _DEBUG_SURFGEO_
	printf("pcoord : %lf,%lf \n",pcoord[iPoint],pcoord[iPoint+1]);
#endif

      /* Upper Left Point ? */
      if (pcoord[iPoint] == minX) 
          if (pcoord[iPoint+1] < minXminZ)
		{
		 minXminZ = pcoord[iPoint+1];
		 iLeftSurf = iAmbient;
		}

      /* Upper Right Point ? */
      if (pcoord[iPoint] == maxX) 
          if (pcoord[iPoint+1] < maxXminZ)
		{
		 maxXminZ = pcoord[iPoint+1];
		 iRightSurf = iAmbient;
		}
    }

    /* Set Pointer to beginning of surface and # of surface lines*/
   
    if (iLeftSurf > iRightSurf)      	
       numSurfaceBorders = iLeftSurf-iRightSurf;
    else
       numSurfaceBorders = iLeftSurf + numAmbients-iRightSurf;
   
    /* Allocate memory for surface data */

    if(((SurfCoordX = (double *)malloc(sizeof(double)*(numSurfaceBorders+1)))
	 == NULL)||
       ((SurfCoordZ = (double *)malloc(sizeof(double)*(numSurfaceBorders+1)))
	 == NULL)||
       ((SurfVecX =   (double *)malloc(sizeof(double)*numSurfaceBorders))
	 == NULL)||
       ((SurfVecZ =   (double *)malloc(sizeof(double)*numSurfaceBorders))
	 == NULL)||
       ((SurfLen =    (double *)malloc(sizeof(double)*numSurfaceBorders))
	 == NULL)||
       ((SurfSumLen = (double *)malloc(sizeof(double)*(numSurfaceBorders + 1)))
	 == NULL))

    {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for ");
      printf ("surface data!\n\n");
      return (2);
    }

    /* Get coordinates of surface points */

#ifdef _DEBUG_SURFGEO_
      printf("\n\nCoordinates of Surface points:\n");
#endif
    iPoint = 0;
    iLeftSurf ++;
    do
       {
       iLeftSurf--;
       if (iLeftSurf < 0) iLeftSurf = numAmbients-1;
       SurfCoordX[iPoint] = pcoord[(ambientBorder[iLeftSurf]-1)*2];
       SurfCoordZ[iPoint] = pcoord[(ambientBorder[iLeftSurf]-1)*2+1];
#ifdef _DEBUG_SURFGEO_
	printf("%d : %lf,%lf \n",iPoint,SurfCoordX[iPoint],SurfCoordZ[iPoint]);
#endif
       iPoint++;
       }
    while(iLeftSurf != iRightSurf);

    /* Calculate length of surface elements and normalized surface vectors */

    SurfSumLen[0]=0;  	 
    for(iPoint=0;iPoint<numSurfaceBorders;iPoint++)
       {
       dX = SurfCoordX[iPoint+1]-SurfCoordX[iPoint];
       dZ = SurfCoordZ[iPoint+1]-SurfCoordZ[iPoint];
       SurfLen[iPoint] = sqrt(dX*dX + dZ*dZ);
       SurfSumLen[iPoint+1] = SurfSumLen[iPoint]+SurfLen[iPoint];
       SurfVecX[iPoint] = dX/SurfLen[iPoint];
       SurfVecZ[iPoint] = dZ/SurfLen[iPoint];
       }
}

/*** i m s u r f g e t d e p t h 

calculates for given point the depth from the surface and lateral distance 
to the left end of the sample

iError : 0 - OK
	 1 - Point lies in area without solution 

***/

#if !( defined(vms) || defined(__vms) )
void imsurfgetdepth_(
#else
void imsurfgetdepth(
#endif
		  double *x,		/* x coordinate of point */
		  double *z,		/* z coordinate of point */
		  double *depth,	/* depth */
	          double *length,	/* lateral distance */
		  int *iError		/* Error Flag */
		  )

{
int IsDepth;
int iSurfMin,iSurf;
double lSurfMin,l,dX,dZ,d;

IsDepth = 0;
for (iSurf=0;iSurf<numSurfaceBorders;iSurf++)
   {/* calculate fraction parallel to surfaceline */
    dX = *x-SurfCoordX[iSurf];
    dZ = *z-SurfCoordZ[iSurf];
    l  = dX*SurfVecX[iSurf] + dZ*SurfVecZ[iSurf];
    /* point is in 'region' of surfaceline ? */
    if ((l>=0) && (l<=SurfLen[iSurf]))
       {
       d = - dX*SurfVecZ[iSurf] + dZ*SurfVecX[iSurf];
       if (fabs(d) < fabs(*depth) || IsDepth == 0)
 	   {
	   IsDepth = 1;
	   *depth = d;
	   lSurfMin = l;
	   iSurfMin = iSurf;
	   }
	}
    } 		
if (IsDepth == 0)
   {
    *iError = 1;
   }
else
   { 
   *length = SurfSumLen[iSurfMin]+lSurfMin;
    *iError = 0;
    }
}


/*** i m s u r f s e t d e p t h 

calculates for given depth from the surface and lateral distance 
to the left end of the sample the coordinates of the appropriate point
(inverse function to imsurfgetdepth)

iError : 0 - OK
	-1 - lateral distance < 0
	 1 - lateral distance exceeds sample dimensions 

***/

#if !( defined(vms) || defined(__vms) )
void imsurfsetdepth_(
#else
void imsurfsetdepth(
#endif
		  double *x,		/* x coordinate of point */
		  double *z,		/* z coordinate of point */
		  double *depth,	/* depth */
	          double *length,	/* lateral distance */
		  int *iError		/* Error Flag */
		  )

{

int iSurf;
double l;

/* Point exceeds right limit */
if (*length > SurfSumLen[numSurfaceBorders])
   {
   *iError = 2;
   return;
   }

iSurf = numSurfaceBorders-1;
while (iSurf>=0 && SurfSumLen[iSurf] > *length)iSurf --;

/* Point exceeds left limit */
if (iSurf < 0)
   {
   *iError = 1;
   return;
   } 

l = *length - SurfSumLen[iSurf];
*x = SurfCoordX[iSurf] + l * SurfVecX[iSurf] - *depth * SurfVecZ[iSurf];
*z = SurfCoordZ[iSurf] + l * SurfVecZ[iSurf] + *depth * SurfVecX[iSurf];
*iError = 0;
}

#if !( defined(vms) || defined(__vms) )
void imsurfgetcoord_(
#else
void imsurfgetcoord(
#endif
		  double *xcoord,	/* x coordinate of point */
		  double *zcoord,	/* z coordinate of point */
		  int    *numcoord	/* Number of coordinates */
		  )
{
int icoord;

*numcoord = numSurfaceBorders+1;

for(icoord=0;icoord<*numcoord;icoord++)
 {
 xcoord[icoord]=SurfCoordX[icoord];
 zcoord[icoord]=SurfCoordZ[icoord];
 }
}
