/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
/*
                          promis / imsil / src
                          --------------------
                              imtestquad.c

  This file contains the test routines for the discretization of a given
  geometry by a finite QUADTREE.

  AUTHORS:   HS  Hannes Stippel, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HS  1994-02-03  new from scratch
  ===========================================================================*/

#include <stdio.h>
#include <math.h>
#include "imquad.h"
#include "imgetmem.h"
#include "imtime.h"
#include "im2dgeo.h"
#include "imtestquad.h"

/* --------------------------------------------------------------------------*/

void printNode (QuadNode *quadNode)
{
   int i;
   
   printf ("This node (%x):\n", quadNode);
   for (i = 0; i < NUM_QUAD_CHILDREN; i++)
   {
      printf ("\tchild #%d is: ", i);
      if (IS_LEAF (quadNode, i))
	 printf ("LEAF (%x)\n", quadNode->quadChild[i]);
      else
	 printf ("NODE (%x)\n", quadNode->quadChild[i]);
   }
   printf ("\tparent node exists: ");
   if (quadNode->quadParent != NODE_NULL)
      printf ("YES (%x)\n", quadNode->quadParent);
   else
      printf ("NO\n");
}

void printLeaf (QuadLeaf *quadLeaf)
{
   int i;
   
   printf ("This leaf (%x):\n", quadLeaf);
   printf ("\thas type: %d\n", quadLeaf->quadLeafType);
   printf ("\tElement is (%x):\n", quadLeaf->quadLeafElement);
   switch (quadLeaf->quadLeafType)
   {
      case 1:
         printf ("\tElement entirely in region: %d\n",
		 quadLeaf->quadLeafElement.quadLeafElementI->region);
         break;
      case 2:
         printf ("\tElement is separated by line from point %d to %d\n",
		 quadLeaf->quadLeafElement.quadLeafElementII->fromPoint,
		 quadLeaf->quadLeafElement.quadLeafElementII->toPoint);
         printf ("\tOn the left there is region %d and on the right %d\n",
		 quadLeaf->quadLeafElement.quadLeafElementII->leftRegion,
		 quadLeaf->quadLeafElement.quadLeafElementII->rightRegion);
         break;
      case 3:
         printf ("\tPoint %d is in this element.\n",
		 quadLeaf->quadLeafElement.quadLeafElementIII->innerPoint);
         printf ("\t%d lines are emerging from this point.\n",
		 quadLeaf->quadLeafElement.quadLeafElementIII->numRegions);
         printf ("\tThe materials are: ");
         for (i = 0; i <
	      quadLeaf->quadLeafElement.quadLeafElementIII->numRegions; i++)
	    printf ("%d ",
		    quadLeaf->quadLeafElement.quadLeafElementIII->region[i]);
	 printf ("\n");
         printf ("\tThey end at points: ");
         for (i = 0; i <
	      quadLeaf->quadLeafElement.quadLeafElementIII->numRegions; i++)
	    printf ("%d ",
		    quadLeaf->quadLeafElement.quadLeafElementIII->endPoint[i]);
         printf ("\n");
         break;
      default:
         printf ("\n\n ***** ERROR:\n Unknown Type!\n\n");
         break;
   }
}

void testQuadTree (QuadNode *quadNode)
{
   char      dec;
   int       num;
   QuadLeaf *quadLeaf;
   
   printf ("\n\nUSE:\n");
   printf ("\tu ... up one node\n");
   printf ("\td ... down one node\n");
   printf ("\te ... end of viewing the QuadTree\n");
   printf ("\n\n");
   
   printNode (quadNode);
   
   printf ("Your choice: ");
   scanf  ("%s", &dec);
   
   while (dec != 'e' && dec != 'E')
   {
      if (dec == 'd' || dec == 'D')
      {
	 printf ("Which one of the nodes (0 ... 3)? ");
	 scanf  ("%d", &num);
	 while (num < 0 || num >= NUM_QUAD_CHILDREN)
	 {
	    printf ("\n\nERROR:\n\tWrong number, must be in 0 ... 3!\n\n");
	    printf ("Which one of the nodes (0 ... 3)? ");
	    scanf  ("%d", &num);
	 }
	 if (IS_LEAF(quadNode,num))
	 {
	    quadLeaf = quadNode->quadChild[num].quadLeaf;
	    printLeaf (quadLeaf);
	    printNode (quadNode);
	 }
	 else
	 {
	    quadNode = quadNode->quadChild[num].quadNode;
	    printNode (quadNode);
	 }
      }
      else
      {
	 if (dec != 'u' && dec != 'U')
	    printf ("\n\nERROR:\n\tWrong choice, use: u(p), d(own),e(nd)\n\n");
	 else
	 {
	    if (quadNode->quadParent == NODE_NULL)
	       printf ("\n\nERROR:\n\tparent does not exist (top level)!\n\n");
	    else
	    {
	       quadNode = quadNode->quadParent;
	       printNode (quadNode);
	    }
	 }
      }
      
      printf ("Your choice: ");
      scanf  ("%s", &dec);
   }
}


/****************************************************************************
 *                                                                          *
 *                             TESTMAIN                                     *
 *                                                                          *
 ****************************************************************************/

void testQuadMain (
		   QuadNode *quadTree,
		   long      numFaces,
		   long      faceEndIndex[],
		   long      lineIndex[],
		   int       numLines,
		   LineType  pointIndex[],
		   int       borderLine[],
		   long      numPoints,
		   int       pointEndIndex[],
		   int       pointIsInLine[],
		   double    pcoord[],
		   double    rootSideX,
		   double    rootSideZ,
		   double    quadResolutionX,
		   double    quadResolutionZ,
		   double    epsma,
		   double    minX,
		   double    maxX,
		   double    minZ,
		   double    maxZ,
		   Point     rootCenter,
		   int       numAmbients,
		   int       ambientBorder[],
		   double    overshoot)
{
   char      dec;
   int       i, j, iend, itest;
   QuadLeaf *quadLeaf;
   Point     point, newCenter;
   double    newSideX, newSideZ;
   double    directionX, directionZ;
   Point     entry;
   
   printf ("\n\nThe Geometry definition\n");
   printf ("\t# of faces: %d\n", numFaces);
   for (i = j = 0; i < numFaces; i++)
   {
      printf ("\tFace %d: # of Lines=%d\n\t\t", i+1, faceEndIndex[i]-j);
      for (; j < faceEndIndex[i]; j++)
	 printf ("%d, ",lineIndex[j]);
      printf ("\n");
   }
   printf ("\t# of points: %d\n", numPoints);
   for (i = 0; i < numPoints; i++)
      printf ("\t\t%g %g\n", pcoord[2*i], pcoord[2*i+1]);
   printf ("\n\n");
   
   printf ("Max. # of Lines: %d\n\n", numLines);
   
   printf ("\n\nThe generated Lines:\n");
   printf ("\t # of Lines: %d\n", numLines);
   for (i = 0; i < numLines; i++)
      printf ("\t\tLine from %d to %d is border between %d and %d\n",
	      pointIndex[i].from, pointIndex[i].to,
	      borderLine[2*i], borderLine[2*i+1]);
   
   printf ("The point information:");
   for (i = j = iend = 0; i < pointEndIndex[numPoints-1]; i++)
   {
      if (i >= iend)
      {
	 printf("\n\t# of lines with point %d: %d\n\t\t",
		j+1, pointEndIndex[j]-iend);
	 iend = pointEndIndex[j];
	 j++;
      }
      printf ("%d ", pointIsInLine[i]);
   }
   
   printf ("\n\n");
   
   printf ("\n\n\tAmbient:\n\t");
   for (i = 0; i < numAmbients; i++)
      printf ("%d, ", ambientBorder[i]);
   
   printf ("\n\nAccuracies and so on:\n");
   printf ("\tEpsma: %g\n", epsma);
   printf ("\tOvershoot: %g\n", overshoot);
   printf ("\tResolution x: %g, Resolution z: %g\n",
	   quadResolutionX, quadResolutionZ);
   printf ("\nThe root rectangle:\n");
   printf ("\tCenter: (%g/%g)\n", rootCenter.x, rootCenter.z);
   printf ("\tSidelength x: %g, Sidelength z: %g\n", rootSideX, rootSideZ);
   printf ("\tmin. x: %g. max. x: %g, min. z: %g. max. z: %g",
	   minX, maxX, minZ, maxZ);
   
   testQuadTree (quadTree);
   
   printf ("\n\nTest the point location.\n");
   
   printf ("Testpoint (y/n): ");
   scanf  ("%s", &dec);
   
   while (dec != 'n' && dec != 'N')
   {
      printf ("X-Coordinate, testpoint: ");
      scanf  ("%lf", &(point.x));
      printf ("Z-Coordinate, testpoint: ");
      scanf  ("%lf", &(point.z));
      printf ("X-Coordinate, testpoint: %g\n", point.x);
      printf ("Z-Coordinate, testpoint: %g\n", point.z);
      
      printf ("Point is in region <%d>\n",
	      getNewRegion (point, quadTree, &quadLeaf, rootCenter, &newCenter,
			    rootSideX, rootSideZ, &newSideX, &newSideZ,
			    pcoord));
      printf ("New Leaf: %x\n", quadLeaf);
      printLeaf (quadLeaf);
      printf ("New center: (%g/%g)\n", newCenter.x, newCenter.z);
      printf ("New sidelength x: %g, z: %g\n", newSideX, newSideZ);
      
      printf ("Testpoint (y/n): ");
      scanf  ("%s", &dec);
   }
   
   printf ("\n\nTest the point location considering old leaf.\n");
   
   printf ("Testpoint (y/n): ");
   scanf  ("%s", &dec);
   
   while (dec != 'n' && dec != 'N')
   {
      printf ("X-Coordinate, testpoint: ");
      scanf  ("%lf", &(point.x));
      printf ("Z-Coordinate, testpoint: ");
      scanf  ("%lf", &(point.z));
      printf ("X-Coordinate, testpoint: %g\n", point.x);
      printf ("Z-Coordinate, testpoint: %g\n", point.z);
      
      printf ("Old Leaf: %x\n", quadLeaf);
      printf ("Point is in region <%d>\n",
	      getRegion (point, quadTree, quadLeaf, &quadLeaf, rootCenter,
			 &newCenter, rootSideX, rootSideZ, &newSideX,
			 &newSideZ, pcoord));
      printf ("New Leaf: %x\n", quadLeaf);
      printLeaf (quadLeaf);
      printf ("New center: (%g/%g)\n", newCenter.x, newCenter.z);
      printf ("New sidelength x: %g, z: %g\n", newSideX, newSideZ);
      
      printf ("Testpoint (y/n): ");
      scanf  ("%s", &dec);
   }
   
   printf ("\n\nTest the determination of the entrance point.\n");
   
   printf ("Test entry (y/n): ");
   scanf  ("%s", &dec);
   
   while (dec != 'n' && dec != 'N')
   {
      printf ("X-Coordinate, testpoint: ");
      scanf  ("%lf", &(point.x));
      printf ("Z-Coordinate, testpoint: ");
      scanf  ("%lf", &(point.z));
      printf ("X-Component of motion: ");
      scanf  ("%lf", &(directionX));
      printf ("Z-Component of motion: ");
      scanf  ("%lf", &(directionZ));
      printf ("X-Coordinate, testpoint: %g\n", point.x);
      printf ("Z-Coordinate, testpoint: %g\n", point.z);
      printf ("X-Component of motion: %g\n", directionX);
      printf ("Z-Component of motion: %g\n", directionZ);
      
      printf ("Ion enters region number <%d>.\n\n",
	      getEntryPoint (point, directionX, directionZ, numAmbients,
			     ambientBorder, pcoord, quadTree, &quadLeaf,
			     rootCenter, &newCenter, rootSideX, rootSideZ,
			     &newSideX, &newSideZ, overshoot, &entry));
      printf ("New Leaf: %x\n", quadLeaf);
      printLeaf (quadLeaf);
      printf ("New center: (%g/%g)\n", newCenter.x, newCenter.z);
      printf ("New sidelength x: %g, z: %g\n", newSideX, newSideZ);
      printf ("Entry point: (%g/%g)\n", entry.x, entry.z);
      
      printf ("Test entry (y/n): ");
      scanf  ("%s", &dec);
   }
   
   printf ("\n\n");
}
