/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
/*
                          promis / imsil / src
                          --------------------
                             imvisualquad.c

  This file contains routines to visualize the finite QUADTREE.

  AUTHORS:   HS  Hannes Stippel, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HS  1994-02-25  new from scratch
  ===========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "imquad.h"
#include "imgetmem.h"
#include "imtime.h"
#include "im2dgeo.h"
#include "imvisualquad.h"
#include "xpgp.h"
/*** declaration of t r n x y z  coded in imf2c.c
     Transforms simulation area coordinates to pixel coordinates ***/

#if defined(__hpux) || defined(vms) || defined(__vms)
void trnxyz  (
#else
void trnxyz_ (
#endif

              int *iloc,                /* locout number */
              double *x,                /* x coordinate in A */
              double *y,                /* y coordinate in A */
              double *z,                /* z coordinate in A */
              double *xx,                       /* x coordinate in pixel*/
              double *xy);                       /* y coordinate in pixel*/

/* --------------------------------------------------------------------------*/

/*** g e t N u m C u r v e s
     determines the number of polygons which will be written to the file. ***/

int getNumCurves (
		  QuadNode *quadNode,	/* quadtree */
		  int       numRegions,	/* # of regions to be visualized */
		  int       region[],	/* regions to be visualized */
		  int       fromLevel[],/* lower level for visualization */
		  int       toLevel[],	/* upper level for visualization */
		  int      *level,	/* level of the quadtree */
		  int      *numPolys)	/* number of polygons */
{
   int                 iChild,			/* i-th child of the node */
                       iRegion,			/* region counter */
                       iLine,			/* line counter for type III */
                       iReturn;			/* return value */
   QuadNode           *newNode;			/* pointer to the next node */
   QuadLeaf           *quadLeaf;		/* pointer to the leaf */
   QuadLeafElementI   *quadLeafElementI;	/* pointer to leaf type I */
   QuadLeafElementII  *quadLeafElementII;	/* pointer to leaf type II */
   QuadLeafElementIII *quadLeafElementIII;	/* pointer to leaf type III */
   
   /* increment the level counter */
   (*level)++;
   
   /* loop over all subnodes of the tree */
   for (iChild = 0; iChild < NUM_QUAD_CHILDREN; iChild++)
   {
      if (IS_LEAF (quadNode, iChild))
      {
	 /* found a leaf!!! */
	 quadLeaf = quadNode->quadChild[iChild].quadLeaf;
	 switch (quadLeaf->quadLeafType)
	 {
	    case 1:
	       /* leaf type I */
	       quadLeafElementI =
		  quadLeaf->quadLeafElement.quadLeafElementI;
	       /* test, if the region shall be recorded */
	       for (iRegion = 0; iRegion < numRegions; iRegion++)
		  if (region[iRegion] == quadLeafElementI->region)
		     if (fromLevel[iRegion] == 0 ||
			 (fromLevel[iRegion] <= *level &&
			  *level <= toLevel[iRegion]))
		     {
			(*numPolys)++;
			break;
		     }
	       break;
	    
	    case 2:
	       /* leaf type I */
	       quadLeafElementII =
		  quadLeaf->quadLeafElement.quadLeafElementII;
	       for (iRegion = 0; iRegion < numRegions; iRegion++)
		  if (region[iRegion] == quadLeafElementII->leftRegion ||
		      region[iRegion] == quadLeafElementII->rightRegion)
		     if (fromLevel[iRegion] == 0 ||
			 (fromLevel[iRegion] <= *level &&
			  *level <= toLevel[iRegion]))
			(*numPolys)++;
	       break;
	    
	    case 3:
	       /* leaf type I */
	       quadLeafElementIII =
		  quadLeaf->quadLeafElement.quadLeafElementIII;
	       for (iRegion = 0; iRegion < numRegions; iRegion++)
		  for (iLine = 0; iLine < quadLeafElementIII->numRegions;
		       iLine++)
		     if (region[iRegion] == quadLeafElementIII->region[iLine])
			if (fromLevel[iRegion] == 0 ||
			    (fromLevel[iRegion] <= *level &&
			     *level <= toLevel[iRegion]))
			   (*numPolys)++;
	       break;
	    
	    default:
	       /* what's that ??? */
	       printf ("\n\n***** ERROR:\n\t");
	       printf ("Unknown leaf type <%d>!\n\n",
		       quadLeaf->quadLeafType);
	       (*level)--;
	       return (-1);
	       break;
	 }
      }
      else
      {
	 newNode = quadNode->quadChild[iChild].quadNode;
	 iReturn = getNumCurves (newNode, numRegions, region, fromLevel,
				 toLevel, level, numPolys);
	 if (iReturn != 0)
	 {
	    (*level)--;
	    return (iReturn);
	 }
      }
   }
   
   (*level)--;
   return (0);
}

/*** g e t P o l y g o n I I
     determines the closed polygon of a leaf type II, which is written to
     the output file. ***/

void getPolygonII (
		   Point   center,		/* center of the leaf */
		   double  sideX,		/* side length x of leaf */
		   double  sideZ,		/* side length z of leaf */
		   Point   beginPoint,		/* begin of line */
		   Point   endPoint,		/* end of line */
		   int     left,		/* region is on left side */
		   int    *numPolyPoints,	/* number of polygon points */
		   Point   polygonPoint[])	/* polygon points */
{
   double vectorX, vectorZ,	/* x/z component of vector of the line */
          xMin, xMax,		/* min/max x coordiante of the leaf */
          zMin, zMax,		/* min/max z coordiante of the leaf */
          lambda, testX,	/* intersection points in line coordinates */
          lambdaMin;		/* minimum distance of intersection point */
   int    iSide1, iSide2,	/* first second intersection found */
          iLambdaMin,		/* intersection with minimum lambda */
          numCorners,		/* number of corners between intersections */
          iCorner,		/* counter for corners between intersections */
          iBegin;		/* first one of the lines for the corners */
   
   /* set the minimum and maximum values of the leaf */
   xMin = center.x - sideX / 2.0;
   xMax = center.x + sideX / 2.0;
   zMin = center.z - sideZ / 2.0;
   zMax = center.z + sideZ / 2.0;
   
   /* get the vector of the line */
   vectorX = endPoint.x - beginPoint.x;
   vectorZ = endPoint.z - beginPoint.z;
   
   /* determine the intersections
      =========================== */
   
   *numPolyPoints = 0;
   
   /* first the horizontal lines */
   if (vectorZ != 0.0)
   {
      /* test line 1 (zMax) */
      lambda = (zMax - beginPoint.z) / vectorZ;
      testX = beginPoint.x + lambda * vectorX;
      if (xMin <= testX && testX <= xMax)
      {
	 /* found an intersection */
	 (*numPolyPoints)++;
	 lambdaMin = lambda;
	 iSide1 = 1;
	 iLambdaMin = 1;
      }
      
      /* test line 3 (zMin) */
      lambda = (zMin - beginPoint.z) / vectorZ;
      testX = beginPoint.x + lambda * vectorX;
      if (xMin <= testX && testX <= xMax)
      {
	 /* found an intersection */
	 if (*numPolyPoints == 0)
	 {
	    lambdaMin = lambda;
	    iSide1 = 3;
	    iLambdaMin = 1;
	 }
	 else
	 {
	    if (lambda < lambdaMin)
	    {
	       /* new minimum intersection distance */
	       testX = lambdaMin;
	       lambdaMin = lambda;
	       lambda = testX;
	       iLambdaMin = 2;
	    }
	    iSide2 = 3;
	 }
	 (*numPolyPoints)++;
      }
   }

   /* the the vertikal lines */
   if (*numPolyPoints < 2 && vectorX != 0.0)
   {
      /* test line 2 (xMin) */
      lambda = (xMin - beginPoint.x) / vectorX;
      testX = beginPoint.z + lambda * vectorZ;
      if (zMin <= testX && testX <= zMax)
      {
	 /* found an intersection */
	 if (*numPolyPoints == 0)
	 {
	    lambdaMin = lambda;
	    iSide1 = 2;
	    iLambdaMin = 1;
	 }
	 else
	 {
	    if (lambda < lambdaMin)
	    {
	       /* new minimum intersection distance */
	       testX = lambdaMin;
	       lambdaMin = lambda;
	       lambda = testX;
	       iLambdaMin = 2;
	    }
	    iSide2 = 2;
	 }
	 (*numPolyPoints)++;
      }
      
      if (*numPolyPoints < 2)
      {
	 /* test line 4 (xMax) */
	 lambda = (xMax - beginPoint.x) / vectorX;
	 testX = beginPoint.z + lambda * vectorZ;
	 if (zMin <= testX && testX <= zMax)
	 {
	    /* found an intersection */
	    if (lambda < lambdaMin)
	    {
	       /* new minimum intersection distance */
	       testX = lambdaMin;
	       lambdaMin = lambda;
	       lambda = testX;
	       iLambdaMin = 2;
	    }
	    iSide2 = 4;
	    (*numPolyPoints)++;
	 }
      }
   }
   
   /* now write the points to the array
      ================================= */
   
   if (left == 1)
   {
      polygonPoint[0].x = beginPoint.x + lambdaMin * vectorX;
      polygonPoint[0].z = beginPoint.z + lambdaMin * vectorZ;
      if (iLambdaMin == 1)
      {
	 iBegin = iSide1;
	 numCorners = iSide2 - iSide1;
      }
      else
      {
	 iBegin = iSide2;
	 numCorners = iSide1 - iSide2;
      }
   }
   else
   {
      polygonPoint[0].x = beginPoint.x + lambda * vectorX;
      polygonPoint[0].z = beginPoint.z + lambda * vectorZ;
      if (iLambdaMin == 1)
      {
	 iBegin = iSide2;
	 numCorners = iSide1 - iSide2;
      }
      else
      {
	 iBegin = iSide1;
	 numCorners = iSide2 - iSide1;
      }
   }
   
   /* this is, when we have a transition from 4 --> ? */
   if (numCorners < 0)
      numCorners += 4;
      
   /* the corner points: always the endpoint of the line */
   for (iCorner = 0; iCorner < numCorners; iCorner++)
      switch ((iCorner+iBegin)%4)
      {
         case 1:
	    /* line 1 (Zmax) */
	    polygonPoint[iCorner+1].x = xMin;
	    polygonPoint[iCorner+1].z = zMax;
	    break;
	    
         case 2:
	    /* line 2 (Xmin) */
	    polygonPoint[iCorner+1].x = xMin;
	    polygonPoint[iCorner+1].z = zMin;
	    break;
	    
         case 3:
	    /* line 3 (Zmin) */
	    polygonPoint[iCorner+1].x = xMax;
	    polygonPoint[iCorner+1].z = zMin;
	    break;
	    
         case 0:
	    /* line 4 (Xmax) */
	    polygonPoint[iCorner+1].x = xMax;
	    polygonPoint[iCorner+1].z = zMax;
	    break;
      }
   
   /* set the number of points in the array */
   *numPolyPoints += numCorners;
   
   /* the last point */
   if (left == 1)
   {
      polygonPoint[numCorners+1].x = beginPoint.x + lambda * vectorX;
      polygonPoint[numCorners+1].z = beginPoint.z + lambda * vectorZ;
   }
   else
   {
      polygonPoint[numCorners+1].x = beginPoint.x + lambdaMin * vectorX;
      polygonPoint[numCorners+1].z = beginPoint.z + lambdaMin * vectorZ;
   }
}

/*** g e t P o l y g o n I I I
     determines the closed polygon of a leaf type III, which is written to
     the output file. ***/

void getPolygonIII (
		    Point   center,		/* center of the leaf */
		    double  sideX,		/* side length x of leaf */
		    double  sideZ,		/* side length z of leaf */
		    Point   firstPoint,		/* begin of line */
		    Point   secondPoint,	/* end of line */
		    Point   innerPoint,		/* point inside leaf */
		    int    *numPolyPoints,	/* number of polygon points */
		    Point   polygonPoint[])	/* polygon points */
{
   double vectorX[2],	/* x component of vectors of the line */
          vectorZ[2],	/* z component of vectors of the line */
          xMin, xMax,	/* min/max x coordiante of the leaf */
          zMin, zMax,	/* min/max z coordiante of the leaf */
          lambda[2],	/* intersection points in line coordinates */
          testX;	/* minimum distance of intersection point */
   int    iSide[2],	/* first second intersection found */
          numCorners,	/* number of corners between intersections */
          iCorner,	/* counter for corners between intersections */
          iCount;	/* counter */
   Point  point[2];	/* endpoints  of the lines */
   
   /* set the minimum and maximum values of the leaf */
   xMin = center.x - sideX / 2.0;
   xMax = center.x + sideX / 2.0;
   zMin = center.z - sideZ / 2.0;
   zMax = center.z + sideZ / 2.0;
   
   /* get the vectors of the lines */
   vectorX[0] = firstPoint.x - innerPoint.x;
   vectorZ[0] = firstPoint.z - innerPoint.z;
   vectorX[1] = secondPoint.x - innerPoint.x;
   vectorZ[1] = secondPoint.z - innerPoint.z;
   
   /* set the endpoints of the lines */
   point[0].x = firstPoint.x;
   point[0].z = firstPoint.z;
   point[1].x = secondPoint.x;
   point[1].z = secondPoint.z;

   /* initialize the pointers to the intersecting lines */
   iSide[0] = iSide[1] = 0;
   
   /* determine the intersections
      =========================== */
   
   *numPolyPoints = 0;
   
   for (iCount = 0; iCount < 2; iCount++)
   {
      if (vectorZ[iCount] != 0.0)
      {
	 /* test line 1 (zMax) */
	 lambda[iCount] = (zMax - innerPoint.z) / vectorZ[iCount];
	 if (lambda[iCount] > 0.0)
	 {
	    testX = innerPoint.x + lambda[iCount] * vectorX[iCount];
	    if (xMin <= testX && testX <= xMax)
	    {
	       /* found an intersection */
	       (*numPolyPoints)++;
	       iSide[iCount] = 1;
	    }
	 }
	 
	 if (iSide[iCount] == 0)
	 {
	    /* test line 3 (zMin) */
	    lambda[iCount] = (zMin - innerPoint.z) / vectorZ[iCount];
	    if (lambda[iCount] > 0.0)
	    {
	       testX = innerPoint.x + lambda[iCount] * vectorX[iCount];
	       if (xMin <= testX && testX <= xMax)
	       {
		  /* found an intersection */
		  (*numPolyPoints)++;
		  iSide[iCount] = 3;
	       }
	    }
	 }
      }
      
      /* the the vertikal lines */
      if (vectorX[iCount] != 0.0 && iSide[iCount] == 0)
      {
	 /* test line 2 (xMin) */
	 lambda[iCount] = (xMin - innerPoint.x) / vectorX[iCount];
	 if (lambda[iCount] > 0.0)
	 {
	    testX = innerPoint.z + lambda[iCount] * vectorZ[iCount];
	    if (zMin <= testX && testX <= zMax)
	    {
	       /* found an intersection */
	       (*numPolyPoints)++;
	       iSide[iCount] = 2;
	    }
	 }
	 
	 if (iSide[iCount] == 0)
	 {
	    /* test line 4 (xMax) */
	    lambda[iCount] = (xMax - innerPoint.x) / vectorX[iCount];
	    if (lambda[iCount] > 0.0)
	    {
	       testX = innerPoint.z + lambda[iCount] * vectorZ[iCount];
	       if (zMin <= testX && testX <= zMax)
	       {
		  /* found an intersection */
		  (*numPolyPoints)++;
		  iSide[iCount] = 4;
	       }
	    }
	 }
      }
   }

   /* this is only, when no intersection with lambda > 0.0 was found.
      Sorry, but this is my first guess! */
   for (iCount = 0; iCount < 2; iCount++)
      if (iSide[iCount] == 0)
      {
	 if (vectorZ[iCount] != 0.0)
	 {
	    /* test line 1 (zMax) */
	    lambda[iCount] = (zMax - innerPoint.z) / vectorZ[iCount];
	    if (lambda[iCount] >= 0.0)
	    {
	       testX = innerPoint.x + lambda[iCount] * vectorX[iCount];
	       if (xMin <= testX && testX <= xMax)
	       {
		  /* found an intersection */
		  (*numPolyPoints)++;
		  iSide[iCount] = 1;
	       }
	    }
	    
	    if (iSide[iCount] == 0)
	    {
	       /* test line 3 (zMin) */
	       lambda[iCount] = (zMin - innerPoint.z) / vectorZ[iCount];
	       if (lambda[iCount] >= 0.0)
	       {
		  testX = innerPoint.x + lambda[iCount] * vectorX[iCount];
		  if (xMin <= testX && testX <= xMax)
		  {
		     /* found an intersection */
		     (*numPolyPoints)++;
		     iSide[iCount] = 3;
		  }
	       }
	    }
	 }
	 
	 /* the the vertikal lines */
	 if (vectorX[iCount] != 0.0 && iSide[iCount] == 0)
	 {
	    /* test line 2 (xMin) */
	    lambda[iCount] = (xMin - innerPoint.x) / vectorX[iCount];
	    if (lambda[iCount] >= 0.0)
	    {
	       testX = innerPoint.z + lambda[iCount] * vectorZ[iCount];
	       if (zMin <= testX && testX <= zMax)
	       {
		  /* found an intersection */
		  (*numPolyPoints)++;
		  iSide[iCount] = 2;
	       }
	    }
	    
	    if (iSide[iCount] == 0)
	    {
	       /* test line 4 (xMax) */
	       lambda[iCount] = (xMax - innerPoint.x) / vectorX[iCount];
	       if (lambda[iCount] >= 0.0)
	       {
		  testX = innerPoint.z + lambda[iCount] * vectorZ[iCount];
		  if (zMin <= testX && testX <= zMax)
		  {
		     /* found an intersection */
		     (*numPolyPoints)++;
		     iSide[iCount] = 4;
		  }
	       }
	    }
	 }
      }
   
   /* now write the points to the array
      ================================= */
   
   /* the first point */
   polygonPoint[0].x = innerPoint.x + lambda[0] * vectorX[0];
   polygonPoint[0].z = innerPoint.z + lambda[0] * vectorZ[0];
   
   numCorners = iSide[1] - iSide[0];
   
   /* this is, when we have a transition from 4 --> ? */
   if (numCorners < 0)
      numCorners += 4;
      
   /* the corner points: always the endpoint of the line */
   for (iCorner = 0; iCorner < numCorners; iCorner++)
      switch ((iCorner+iSide[0])%4)
      {
         case 1:
	    /* line 1 (Zmax) */
	    polygonPoint[iCorner+1].x = xMin;
	    polygonPoint[iCorner+1].z = zMax;
	    break;
	    
         case 2:
	    /* line 2 (Xmin) */
	    polygonPoint[iCorner+1].x = xMin;
	    polygonPoint[iCorner+1].z = zMin;
	    break;
	    
         case 3:
	    /* line 3 (Zmin) */
	    polygonPoint[iCorner+1].x = xMax;
	    polygonPoint[iCorner+1].z = zMin;
	    break;
	    
         case 0:
	    /* line 4 (Xmax) */
	    polygonPoint[iCorner+1].x = xMax;
	    polygonPoint[iCorner+1].z = zMax;
	    break;
      }
   
   /* set the number of points in the array */
   *numPolyPoints += numCorners;
   
   /* the last point */
   polygonPoint[numCorners+1].x = innerPoint.x + lambda[1] * vectorX[1];
   polygonPoint[numCorners+1].z = innerPoint.z + lambda[1] * vectorZ[1];
}

/*** w r i t e Q u a d T r e e
     writes the quadtree to the output file. ***/

int writeQuadTree (
		   FILE     *quadFile,		/* output file */
		   QuadNode *quadNode,		/* quadtree */
		   Point     center,		/* center of the node */
		   double    sideX,		/* side of the node, x */
		   double    sideZ,		/* side of the node, z */
		   double    pcoord[],		/* coordinates of the points */
		   double    xscale,		/* geometry scaling factor */
		   int       numRegions,	/* # of regions to visualize */
		   int       region[],		/* regions to be visualized */
		   int       fromLevel[],	/* lower level */
		   int       toLevel[],		/* upper level */
		   int      *level)		/* level of the quadtree */
{
   Point               newCenter,		/* new center point */
                       firstPoint,		/* first point of line */
                       secondPoint,		/* second point of line */
                       innerPoint,		/* inner point of leaf III */
                       polygonPoint[5];		/* points of the polygon */
   double              newSideX,		/* new sidelength, x */
                       newSideZ;		/* new sidelength, z */
   int                 iChild,			/* i-th child of the node */
                       iRegion,			/* region counter */
                       iLine,			/* line counter for type III */
                       iPoint,			/* point index */
                       left,			/* indicates left region */
                       numPolyPoints,		/* # of points of polygon */
                       iPolyPoint,		/* counter for polygonpoints */
                       iReturn;			/* return value */
   QuadNode           *newNode;			/* pointer to the next node */
   QuadLeaf           *quadLeaf;		/* pointer to the leaf */
   QuadLeafElementI   *quadLeafElementI;	/* pointer to leaf type I */
   QuadLeafElementII  *quadLeafElementII;	/* pointer to leaf type II */
   QuadLeafElementIII *quadLeafElementIII;	/* pointer to leaf type III */
   
   /* increment the level counter */
   (*level)++;
   
   /* determine new side length's */
   newSideX = sideX / 2.0;
   newSideZ = sideZ / 2.0;
   
   /* loop over all subnodes of the tree */
   for (iChild = 0; iChild < NUM_QUAD_CHILDREN; iChild++)
   {
      /* determine new center */
      newCenter.x = NEW_X (iChild, center.x, newSideX/2.0);
      newCenter.z = NEW_Z (iChild, center.z, newSideZ/2.0);
      
      if (IS_LEAF (quadNode, iChild))
      {
	 /* found a leaf!!! */
	 quadLeaf = quadNode->quadChild[iChild].quadLeaf;
	 switch (quadLeaf->quadLeafType)
	 {
	    case 1:
	       /* leaf type I */
	       quadLeafElementI =
		  quadLeaf->quadLeafElement.quadLeafElementI;
	       /* test, if the region shall be recorded */
	       for (iRegion = 0; iRegion < numRegions; iRegion++)
		  if (region[iRegion] == quadLeafElementI->region)
		     if (fromLevel[iRegion] == 0 ||
			 (fromLevel[iRegion] <= *level &&
			  *level <= toLevel[iRegion]))
		     {
			/* write the rectangle to the file */
			fprintf (quadFile, "5\n");
			fprintf (quadFile, "%f %f\n",
				 (newCenter.x-newSideX/2.0)*xscale,
				 (newCenter.z-newSideZ/2.0)*xscale*(-1.0));
			fprintf (quadFile, "%f %f\n",
				 (newCenter.x-newSideX/2.0)*xscale,
				 (newCenter.z+newSideZ/2.0)*xscale*(-1.0));
			fprintf (quadFile, "%f %f\n",
				 (newCenter.x+newSideX/2.0)*xscale,
				 (newCenter.z+newSideZ/2.0)*xscale*(-1.0));
			fprintf (quadFile, "%f %f\n",
				 (newCenter.x+newSideX/2.0)*xscale,
				 (newCenter.z-newSideZ/2.0)*xscale*(-1.0));
			fprintf (quadFile, "%f %f\n",
				 (newCenter.x-newSideX/2.0)*xscale,
				 (newCenter.z-newSideZ/2.0)*xscale*(-1.0));
			break;
		     }
	       break;
	    
	    case 2:
	       /* leaf type I */
	       quadLeafElementII =
		  quadLeaf->quadLeafElement.quadLeafElementII;
	       for (iRegion = 0; iRegion < numRegions; iRegion++)
		  if (region[iRegion] == quadLeafElementII->leftRegion ||
		      region[iRegion] == quadLeafElementII->rightRegion)
		     if (fromLevel[iRegion] == 0 ||
			 (fromLevel[iRegion] <= *level &&
			  *level <= toLevel[iRegion]))
		     {
			/* determine the points of the polygon */
			iPoint = 2 * (quadLeafElementII->fromPoint - 1);
			firstPoint.x = pcoord[iPoint];
			firstPoint.z = pcoord[iPoint+1];
			iPoint = 2 * (quadLeafElementII->toPoint - 1);
			secondPoint.x = pcoord[iPoint];
			secondPoint.z = pcoord[iPoint+1];
			if (region[iRegion] == quadLeafElementII->leftRegion)
			   left = 1;
			else
			   left = 0;
			getPolygonII (newCenter, newSideX, newSideZ,
				      firstPoint, secondPoint, left,
				      &numPolyPoints, polygonPoint);
			/* write the polygon to the output file */
			fprintf (quadFile, "%d\n", numPolyPoints+1);
			for (iPoint = 0; iPoint < numPolyPoints; iPoint++)
			   fprintf (quadFile, "%f %f\n",
				    polygonPoint[iPoint].x*xscale,
				    polygonPoint[iPoint].z*xscale*(-1.0));
			fprintf (quadFile, "%f %f\n",
				 polygonPoint[0].x*xscale,
				 polygonPoint[0].z*xscale*(-1.0));
		     }
	       break;
	    
	    case 3:
	       /* leaf type I */
	       quadLeafElementIII =
		  quadLeaf->quadLeafElement.quadLeafElementIII;
	       for (iRegion = 0; iRegion < numRegions; iRegion++)
		  for (iLine = 0; iLine < quadLeafElementIII->numRegions;
		       iLine++)
		     if (region[iRegion] == quadLeafElementIII->region[iLine])
			if (fromLevel[iRegion] == 0 ||
			    (fromLevel[iRegion] <= *level &&
			     *level <= toLevel[iRegion]))
			{
			   /* determine the points of the polygon */
			   iPoint =
			      2 * (quadLeafElementIII->endPoint[iLine] - 1);
			   firstPoint.x = pcoord[iPoint];
			   firstPoint.z = pcoord[iPoint+1];
			   if (iLine == quadLeafElementIII->numRegions - 1)
			      iPoint =
				 2 * (quadLeafElementIII->endPoint[0] - 1);
			   else
			      iPoint = 2 *
				 (quadLeafElementIII->endPoint[iLine+1] - 1);
			   secondPoint.x = pcoord[iPoint];
			   secondPoint.z = pcoord[iPoint+1];
			   iPoint = 
			      2 * (quadLeafElementIII->innerPoint - 1);
			   innerPoint.x = pcoord[iPoint];
			   innerPoint.z = pcoord[iPoint+1];
			   getPolygonIII (newCenter, newSideX, newSideZ,
					  firstPoint, secondPoint, innerPoint,
					  &numPolyPoints, polygonPoint);
			   /* write the polygon to the output file */
			   fprintf (quadFile, "%d\n", numPolyPoints+2);
			   fprintf (quadFile, "%f %f\n",
				    innerPoint.x*xscale,
				    innerPoint.z*xscale*(-1.0));
			   for (iPoint = 0; iPoint < numPolyPoints; iPoint++)
			      fprintf (quadFile, "%f %f\n",
				       polygonPoint[iPoint].x*xscale,
				       polygonPoint[iPoint].z*xscale*(-1.0));
			   fprintf (quadFile, "%f %f\n",
				    innerPoint.x*xscale,
				    innerPoint.z*xscale*(-1.0));
			}
	       break;
	    
	    default:
	       /* what's that ??? */
	       printf ("\n\n***** ERROR:\n\t");
	       printf ("Unknown leaf type <%d>!\n\n",
		       quadLeaf->quadLeafType);
	       (*level)--;
	       return (-1);
	       break;
	 }
      }
      else
      {
	 newNode = quadNode->quadChild[iChild].quadNode;
	 iReturn = writeQuadTree (quadFile, newNode, newCenter, newSideX,
				  newSideZ, pcoord, xscale, numRegions,
				  region, fromLevel, toLevel, level);
	 if (iReturn != 0)
	 {
	    (*level)--;
	    return (iReturn);
	 }
      }
   }
   
   (*level)--;
   return (0);
}

/*** v i s u a l Q u a d M a i n
     is the main routine for the visualization of the quadtree. ***/

int visualQuadMain (
		    QuadNode *quadTree,		/* quadtree */
		    Point     rootCenter,	/* center of the root node */
		    double    rootSideX,	/* side of the root node, x */
		    double    rootSideZ,	/* side of the root node, z */
		    double    minX,		/* minimum x coordinate */
		    double    maxX,		/* maximum x coordinate */
		    double    minZ,		/* minimum z coordinate */
		    double    maxZ,		/* maximum z coordinate */
		    long      numFaces,		/* number of regions */
		    double    pcoord[],		/* coordinates of the points */
		    double    xscale)		/* geometry scaling factor */
{
   int   iCount = 0,		/* counter */
         iRegion,		/* region counter */
         numRegions,		/* number of regions to be visualized */
        *region,		/* regions to be visualized */
        *fromLevel,		/* lower level for visualization */
        *toLevel,		/* upper level for visualization */
         level,			/* level identifier */
         numPolys;		/* number of polygons */
   char  quadFileName[16];	/* name of the output file */
   FILE *quadFile;		/* output file */
   
   if ((region = (int *) malloc (sizeof (int) * numFaces)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for the ");
      printf ("regions!\n\n");
      return (1);
   }
   if ((fromLevel = (int *) malloc (sizeof (int) * numFaces)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for the ");
      printf ("fromLevels!\n\n");
      return (2);
   }
   if ((toLevel = (int *) malloc (sizeof (int) * numFaces)) == NULL)
   {
      printf ("\n\n\t***** ERROR:\n\tCould not allocate memory for the ");
      printf ("toLevels!\n\n");
      return (3);
   }
   
   printf ("\n\nNumber of regions to be visualized (<= 0 for none): ");
   scanf ("%d", &numRegions);
   while (numRegions > 0)
   {
      while (numRegions > numFaces)
      {
	 printf ("\n\t THERE ARE ONLY %d REGIONS IN THIS GEOMETRY!\n\n",
		 numFaces);
	 printf ("\nNumber of regions to be visualized (<= 0 for none): ");
	 scanf ("%d", &numRegions);
      }
      if (numRegions > 0)
      {
	 iCount++;
	 for (iRegion = 0; iRegion < numRegions; iRegion++)
	 {
	    printf ("\n\tRegion index: ");
	    scanf  ("%d", &(region[iRegion]));
	    while (region[iRegion] < 1 || region[iRegion] > numFaces)
	    {
	       printf ("\n\t\t THE REGION NUMBER MUST BE BETWEEN 1 AND %d!\n",
		       numFaces);
	       printf ("\n\tRegion index: ");
	       scanf  ("%d", &(region[iRegion]));
	    }
	    printf ("\tBegin at Level (0 for all levels): ");
	    scanf  ("%d", &(fromLevel[iRegion]));
	    while (fromLevel[iRegion] < 0)
	    {
	       printf ("\n\t\t THE LEVEL NUMBER MUST BE LARGER THEN 0!\n");
	       printf ("\tBegin at Level (0 for all levels): ");
	       scanf  ("%d", &(fromLevel[iRegion]));
	    }
	    if (fromLevel[iRegion] > 0)
	    {
	       printf ("\tEnd at Level: ");
	       scanf  ("%d", &(toLevel[iRegion]));
	       while (toLevel[iRegion] < fromLevel[iRegion])
	       {
		  printf ("\n\t\t THE LEVEL NUMBER MUST BE LARGER OR EQUAL ");
		  printf ("THE FROM LEVEL (= %d)!\n", fromLevel[iRegion]);
		  printf ("\tEnd at Level: ");
		  scanf  ("%d", &(toLevel[iRegion]));
	       }
	    }
	    else
	       toLevel[iRegion] = 0;
	 }
	 
	 /* open the appropriate output file */
	 if (iCount < 10)
	    sprintf (quadFileName, "QUADTREE_00%1d.DAT", iCount);
	 else
	    if (iCount < 100)
	       sprintf (quadFileName, "QUADTREE_0%2d.DAT", iCount);
	    else
	       if (iCount < 1000)
		  sprintf (quadFileName, "QUADTREE_%3d.DAT", iCount);
	       else
	       {
		  printf ("\n\n***** ERROR:\n\t");
		  printf ("Can not open anymore files! Continue in ");
		  printf ("Program!\n\n");
		  free (toLevel);
		  free (fromLevel);
		  free (region);
	       }
	 if ((quadFile = fopen (quadFileName,"w")) == NULL)
	 {
	    printf ("\n\n***** ERROR:\n\t");
	    printf ("Could not open File <%s>!\n\n", quadFileName);
	    free (toLevel);
	    free (fromLevel);
	    free (region);
	    return (4);
	 }
	 
	 /* write the header to the file */
	 fprintf (quadFile, "%f %f %f %f\n",
		  minX*xscale, maxX*xscale,
		  maxZ*xscale*(-1.0), minZ*xscale*(-1.0));
	 
	 /* determine the number of polygons and write it to the file */
	 level = 0;
	 numPolys = 0;
	 if (getNumCurves (quadTree, numRegions, region, fromLevel, toLevel,
			   &level, &numPolys) < 0)
	 {
	    printf ("\n\n***** ERROR:\n\t");
	    printf ("During Counting the polygons to be written!\n\n");
	    free (toLevel);
	    free (fromLevel);
	    free (region);
	    return (5);
	 }
	 fprintf (quadFile, "%d\n", numPolys);
	 
	 /* write the quadtree itself to the file */
	 level = 0;
	 if (writeQuadTree (quadFile, quadTree, rootCenter, rootSideX,
			    rootSideZ, pcoord, xscale, numRegions, region,
			    fromLevel, toLevel, &level) < 0)
	 {
	    printf ("\n\n***** ERROR:\n\t");
	    printf ("During writing the polygons to the file!\n\n");
	    free (toLevel);
	    free (fromLevel);
	    free (region);
	    return (6);
	 }
	 
	 fclose (quadFile);
	 printf ("\n\nNumber of regions to be visualized (<= 0 for none): ");
	 scanf ("%d", &numRegions);
      }
   }
   
   free (toLevel);
   free (fromLevel);
   free (region);
   
   return(0);
}

#if defined(__hpux) || defined(vms) || defined(__vms)
#else
#define trnxyz trnxyz_ 
#endif


/*** d r a w Q u a d T r e e
     draws the quadtree in a window. ***/

void drawQuadTree (
		   QuadNode *quadNode,	/* node of the quadtree */
		   Point     center,	/* center of the rectangle */
		   double    sideX,	/* side x of the rectangle */
		   double    sideZ,	/* side z of the rectangle */
      		   int 	     *iloc       /* location dep. output */
                  )
{
   int    iChild;		/* i-th child of the node */
   double VerticalX,		/* x-position of vertical line */
          VerticalFrom,		/* first z-coordinate for vertical line */
          VerticalTo,		/* second z-coordinate for vertical line */
          HorizontalZ,		/* z-position of vertical line */
          HorizontalFrom,	/* first x-coordinate for vertical line */
          HorizontalTo;		/* second x-coordinate for vertical line */
   Point  newCenter;		/* center of the new leaf node */
   double X,Y,Z;		/* X,Y,Z Coordinates */
   double newSideX = sideX/2.0,	/* new side length x */
          newSideZ = sideZ/2.0;	/* new side length z */
   
   /* get the coordinates of the axes */
   X = center.x;
   Y = 0.;
   Z = center.z;
   trnxyz(iloc,&X,&Y,&Z,&VerticalX,&HorizontalZ);

/*   VerticalX      = borderX + (int) ((center.x + offsetX) * scaleX);
   HorizontalZ    = borderZ + (int) ((center.z + offsetZ) * scaleZ);*/

   X = center.x-newSideX;
   Y = 0.;
   Z = center.z-newSideZ;
   trnxyz(iloc,&X,&Y,&Z,&HorizontalFrom,&VerticalFrom);
/*   VerticalFrom   = borderZ + (int) ((center.z-newSideZ + offsetZ) * scaleZ);
   HorizontalFrom = borderX + (int) ((center.x-newSideX + offsetX) * scaleX);
*/

   X = center.x+newSideX;
   Y = 0.;
   Z = center.z+newSideZ;
   trnxyz(iloc,&X,&Y,&Z,&HorizontalTo,&VerticalTo);
/*   VerticalTo     = borderZ + (int) ((center.z+newSideZ + offsetZ) * scaleZ);
   HorizontalTo   = borderX + (int) ((center.x+newSideX + offsetX) * scaleX);
*/
   
   /* test the size of the rectangle */
   if ((VerticalTo-VerticalFrom) > 5. && (HorizontalTo-HorizontalFrom > 5.))
   {
      /* draw the axes for this node/leaf */
      Move   (VerticalX, VerticalFrom);
      LineTo (VerticalX, VerticalTo);
      Move   (HorizontalFrom, HorizontalZ);
      LineTo (HorizontalTo, HorizontalZ);
      
      /* loop over all subnodes of the tree */
      for (iChild = 0; iChild < NUM_QUAD_CHILDREN; iChild++)
      {
	 /* get the new center */
	 newCenter.x = NEW_X (iChild, center.x, newSideX/2.0);
	 newCenter.z = NEW_Z (iChild, center.z, newSideZ/2.0);
	 
	 /* draw the next node */
	 if (!IS_LEAF(quadNode, iChild))
	    drawQuadTree (quadNode->quadChild[iChild].quadNode, newCenter,
			  newSideX,newSideZ,iloc);
      }
   }
}
