/*****************************************************************************/
/*     Copyright (c) 2000, Compaq Computer Corporation                       */
/*                                                                           */
/*  See the file "license.terms" for information on usage and redistribution */
/*  of this file, and for a DISCLAIMER OF ALL WARRANTIES.                    */
/*****************************************************************************/
static char *_rcsid_="$Id: pif2quad.c,v 1.0 1994/03/22 17:00:00 stippel Exp $";
/*
                          promis / imsil / pif
                          --------------------
                               pif2quad.c

  This file contains the program to convert the geometry from a pif file
  to the format required for the input for the generation of the finite
  used in quadtree. Additionally a "frame"-input deck for IMSIL2D is
  generated including the "LAYER" information (see chapter 4 of the IMSIL2D
  reference manual).

  AUTHORS:   HS  Hannes Stippel, CEC Vienna, Digital Equipment Corporation

  who when        what
  ---------------------------------------------------------------------------
  HS  1994-03-22  new from scratch
  ===========================================================================*/

#include "vista.h"
#include "vmath.h"
#include "verrstd.h"
#include "verrP.h"
#include "pai.h"
#include "pal.h"
#include "mat.h"
#include "g2s.h"
#include "gas.h"
#include "vrr.h"
#include "vexit.h"

/* --------------------------------------------------------------------------*/

vrrResourceDef commandParameter[] = {	/* the command line arguments */
/* required parameters */
{ "inputPhysical", "inphy",   vrrREQUIRED_FILE, vNULL,
  "name of the physical input pif file" },
/* optional parameters */
{ "inputLogical",  "inlog",   vrrKEY_STRING,    vNULL,
  "name of the logical input pif file" },
{ "inputGeometry", "ingeo",   vrrKEY_STRING,    vNULL,
  "name of the input pif geometry" },
{ "output",        "outnam",  vrrKEY_STRING,    "GEO.DAT",
  "name of the output file" },
{ "inputDeck",     "decknam", vrrKEY_STRING,    "INP.DAT",
  "name of the input deck file" },
/* terminating table */
{ vNULL }
};

/*paiString elementName = {"Element   "};	/* element name */

/* --------------------------------------------------------------------------*/

/*** r e a d C o m m a n d L i n e
     reads the input parameters from the command line by help of the VISTA
     command line interface. ***/

vVoid readCommandLine (vConstString *physicalName,	/* physical PIF file */
                       vConstString *logicalName,	/* logical PIF file */
                       vConstString *geometryName,	/* PIF geometry */
                       vConstString *outputName,	/* output file name */
                       vConstString *inputDeckName)	/* input deck name */
{
   VeStart ("readCommandLine");
   
   *physicalName  = vrrGetCmdlineString ("inputPhysical");
   *logicalName   = vrrGetCmdlineString ("inputLogical");
   *geometryName  = vrrGetCmdlineString ("inputGeometry");
   *outputName    = vrrGetCmdlineString ("output");
   *inputDeckName = vrrGetCmdlineString ("inputDeck");
   
   VeReturnVoid;
}

/*** e n t e r N a m e
     is an auxiliary function for the function getOnlyLogicalPif. 
     It is used as an argument to the PAI function which finds all the 
     logical PIF files in a physical file and calls a user-defined 
     function with the logical file as an argument. This is the 
     user-defined function for this purpose. It just copies the logical 
     name to a place (this place is given as the user-defined data). ***/

paiBoolean enterName (vosFile        physical,	/* physical PIF file */  
		      paiVoid        *userData,	/* user data */
		      paiConstString logical)	/* logical file name */
{
   paiString *uD = (paiString*) userData;
   
   VeStart("enterName");
   
   if (**uD != '\0')
   {
      **uD = '\0';
      VeReturn(paiTRUE);
   }
   paiStrcpy (*uD, logical);
   
   VeReturn(paiFALSE);
}

/*** g e t O n l y L o g i c a l P i f
     looks for logical PIF files in the specified physical PIF. 
     If it finds exactly one logical file, the name of this file is 
     returned. Otherwise vNULL is returned, even if there are more
     logical pif files. The returned pointer pointer is allocated by 
     paiStrdup, so it must be freed with paiFree sometime in the future. */

paiString getOnlyLogicalPif (vConstString physical)  /* physical PIF file */  
{
   paiNameString userData;			/* user data */
   paiString     userDataP = &(userData[0]);	/* address of user data */
   vosFile       vosFileName;			/* pif file name */
   
   VeStart ("getOnlyLogicalPif");
   
   userData[0] = '\0';
   vosFileName = vosOs2Vospec (physical, vFALSE);
   if (palDirFile(vosFileName, enterName, (paiVoid*)(&userDataP))
       == paiFALSE)
   {
      vosFreeVospec (vosFileName);
      VeReturn (paiNULL);
   }
   
   vosFreeVospec (vosFileName);
   if (userData[0] == '\0')
      VeReturn (paiNULL);
   
   VeReturn (paiStrdup(userData));
}

/*** c o n v e r t G e o m e t r y
     converts the geometry from the g2s geometry format to a format suitable
     for the finite QUADTREE used in IMSIL2D and writes it to the output
     file. ***/

vBoolean convertGeometry (g2sGeometryPtr  g2sGeometry,	/* geometry */
			  FILE           *outputFile)	/* output file */
{
   g2sPointPtr g2sPoint,	/* points of the geometry */
               g2sFirstPoint;	/* first point of a face */
   g2sLinePtr  g2sLine,		/* lines of the geometry */
               g2sPreviousLine,	/* previous line */
               g2sNextLine;	/* next line */
   g2sFacePtr  g2sFace;		/* faces of the geometry */
   vInt        numLines,	/* number of lines of a face */
               iFace,		/* face counter */
               iPoint;		/* point counter */
   vDouble     zCoordinate;	/* z-coordinate of the point */
   
   VeStart ("convertGeometry");
   
   /* the number of faces */
   fprintf (outputFile, "%d\n", G2sNumberOfFaces (g2sGeometry));
   
   /* the first face */
   g2sFace = G2sFirstFace (g2sGeometry);
   
   /* loop over all faces */
   for (iFace = 0; iFace < G2sNumberOfFaces(g2sGeometry); iFace++)
   {
      /* initialize the lines counter for the face */
      numLines = 1;
      
      /* the first line of the face */
      g2sLine = g2sFace->line;
      
      if (g2sLine != vNULL)
      {
	 /* the starting point of this line */
	 if (G2sLineLeftFace (g2sLine) == g2sFace)
	 {
	    g2sFirstPoint = G2sLineEnd   (g2sLine);
	    g2sPoint      = G2sLineStart (g2sLine);
	 }
	 else
	 {
	    if (G2sLineRightFace (g2sLine) == g2sFace)
	    {
	       g2sFirstPoint = G2sLineStart (g2sLine);
	       g2sPoint      = G2sLineEnd   (g2sLine);
	    }
	    else
	    {
	       VeError (veALG, "Line of a face is not border of the face?");
	       VeReturn (vFALSE);
	    }
	 }
	 
	 /* count the lines of this face */
	 do
	 {
	    numLines++;
	    
	    /* get the next line */
	    if (g2sGetNextLines (g2sLine, g2sPoint, &g2sNextLine,
				 &g2sPreviousLine) == vFALSE) 
	    {
               VeError (veALG, "Could not get next line.");
               VeReturn (vFALSE);
	    }
	    g2sLine = g2sNextLine;
	    
	    /* get the next point */
	    g2sPoint = G2sGetOppositeLineEnd (g2sLine, g2sPoint);
	 } while (g2sPoint != g2sFirstPoint);
	 
	 /* write the number of lines of this face to the output file */
	 fprintf (outputFile, "%d\n", numLines);
	 
	 /* write the point references for this face;
	    get the first line again */
	 g2sLine = g2sFace->line;
	 
	 /* write the first point of the face to the output file */
	 fprintf (outputFile, " %d\n", g2sFirstPoint->num);
	 
	 /* get the second point of the face */
	 g2sPoint = G2sGetOppositeLineEnd (g2sLine, g2sFirstPoint);
	 
	 /* loop over the rest of the face */
	 while (g2sPoint != g2sFirstPoint)
	 {
	    /* write the point to the output file */
	    fprintf (outputFile, " %d\n", g2sPoint->num);
	    
	    /* get the next line */
	    g2sGetNextLines (g2sLine, g2sPoint, &g2sNextLine,
			     &g2sPreviousLine);
	    g2sLine = g2sNextLine;
	    
	    /* get the next point */
	    g2sPoint = G2sGetOppositeLineEnd (g2sLine, g2sPoint);
	 }
      }
      
      /* get the next face */
      g2sFace = G2sNextFace (g2sGeometry, g2sFace);
   }
   
   /* write the number of points */
   fprintf (outputFile, "%d\n", G2sNumberOfPoints (g2sGeometry));
   for (iPoint = 1; iPoint <= G2sNumberOfPoints (g2sGeometry); iPoint++)
   {
      /* get the point */
      g2sPoint = G2sPointByIndex (g2sGeometry, iPoint);
      
      /* write it to the output file */
      zCoordinate = 0.0 - 0.0-g2sPoint->x[1];
      fprintf (outputFile, "%lf %lf\n", g2sPoint->x[0], zCoordinate);
   }
   
   /* no grid */
   fprintf (outputFile, "0\n0\n");
   
   VeReturn (vTRUE);
}

/*** c o n v e r t A t t r i b u t e s
     generates the frame input deck for IMSIL2D from the PIF material
     description. ***/

vBoolean convertAttributes (gasAttributesPtr  gasAttributes,	/* materials */
			    g2sGeometryPtr    g2sGeometry,	/* geometry */
			    FILE             *outputFile)	/* inputdeck */
{
   vInt          iMaterial,		/* material counter */
                *segmentNumber,		/* segment number of the face */
                 numAtoms = 0,		/* total number of atoms in the area */
                 iSearch,		/* counter for different atoms */
                 numDifferentAtoms = 0,	/* number of different atom species */
                *iAtom,			/* atomic number */
                *iProbability,		/* probability to encounter an atom */
                 number;		/* material composition number */
   vConstString  matMaterial;		/* material names */
   vString       material,		/* material names */
                *formula;  		/* chemical formula of the material */
   vLong         mode,			/* material comparison mode */
                 atomNumber;		/* atomic number */
   vDouble      *density,		/* density of the materials */
                 atomMass,		/* atomic mass */
                *mass;			/* atomic mass of the atoms */
   vChar        *xtal,			/* info, about crystalline material */
                cAtom[3],		/* atom name */
                elementName[9];		/* element names */
   g2sSgmnt_t  *segment;		/* segment */
   g2sFace_t   *face;			/* face */
   
   VeStart ("convertAttributes");

   /* terminate the atom name */
   cAtom[2] = '\0';
   
   /* allocate the memory for the density and the material info */
   if ((segmentNumber = (vInt *) vMalloc (sizeof (vInt) *
					  G2sNumberOfFaces(g2sGeometry))) ==
       vNULL)
   {
      VeError (veSYS, "Could not allocate enough memory for segment array.");
      VeReturn (vFALSE);
   }
   if ((density = (vDouble *) vMalloc (sizeof (vDouble) *
				       G2sNumberOfSegments(g2sGeometry))) ==
       vNULL)
   {
      VeError (veSYS, "Could not allocate enough memory for density array.");
      VeReturn (vFALSE);
   }
   if ((xtal = (vChar *) vMalloc (sizeof (vChar) *
				  G2sNumberOfSegments(g2sGeometry))) == vNULL)
   {
      VeError (veSYS, "Could not allocate enough memory for xtal array.");
      VeReturn (vFALSE);
   }
   if ((formula = (vString *) vMalloc (sizeof (vString) *
				       G2sNumberOfSegments(g2sGeometry))) ==
       vNULL)
   {
      VeError (veSYS, "Could not allocate enough memory for formula array.");
      VeReturn (vFALSE);
   }
   
   for (iMaterial = 0; iMaterial < G2sNumberOfFaces(g2sGeometry); iMaterial++)
   {
      /* get the segment number from the face index */
      face = G2sFaceByIndex (g2sGeometry, iMaterial+1);
      segment = face->sgmnt;
      segmentNumber[iMaterial] = segment->num;
   }
   
   for (iMaterial = 0;
	iMaterial < G2sNumberOfSegments(g2sGeometry);
	iMaterial++)
   {
      /* default is noncrystalline materials */
      xtal[iMaterial] = 'F';
      
      /* get the material name */
      if (gasPifGetString (gasAttributes, iMaterial+1, "",
			   "SegmentDescription", 0, "", "MaterialType", 0,
			   &matMaterial, vFALSE) == vFALSE)
	 VeReturn (vFALSE);
      
      /* check and convert the new material information */
      if (matExists (matMaterial))
      {
	 matCompare (matMaterial, "m-Si", &mode);
	 if (mode >= matSUBCLASS)
	 {
	    /* this is the mono silicon */
            xtal[iMaterial] = 'T';
            
            /* get the memory for the Silicon string itself */
	    if ((formula[iMaterial] =
		 (vChar *) vMalloc (sizeof (vChar) * 3)) == vNULL)
	    {
	       VeError2 (veSYS, "Could not allocate enough memory %d. %s",
			 iMaterial, "for formula");
	       VeReturn (vFALSE);
	    }
	    
            /* set the formula */
            vStrcpy (formula[iMaterial], "Si");
	    density[iMaterial] = 4.977e22;
	 }
	 else
	 {
	    if ((material = (vChar *) vMalloc (sizeof (vChar) *
					       vStrlen(matMaterial))) == vNULL)
	    {
	       VeError2 (veSYS, "Could not allocate enough memory %d. %s",
			 iMaterial, "for material");
	       VeReturn (vFALSE);
	    }
            vStrcpy (material, matMaterial);
            if ((matMaterial = matGetDefaultName (material)) != vNULL)
	    {
	       vFree (material);
	       if ((material = (vChar *)
		    vMalloc (sizeof (vChar) * vStrlen(matMaterial))) == vNULL)
	       {
		  VeError2 (veSYS, "Could not allocate enough memory %d. %s",
			    iMaterial, "for material");
		  VeReturn (vFALSE);
	       }
               vStrcpy (material, matMaterial);
	    }
	    
            matCompare (material, "PMMA", &mode);
            if (mode >= matSUBCLASS)
            {
               /* this is PMMA; get the memory for the formula string */
               if ((formula[iMaterial] =
		    (vChar *) vMalloc (sizeof (vChar) * 7)) == vNULL)
               {
                  VeError2 (veSYS, "Could not allocate enough memory %d. %s",
                            iMaterial, "for formula");
                  VeReturn (vFALSE);
               }
	       
	       /* set the formula */
	       vStrcpy (formula[iMaterial], "C5H8O2");
               density[iMaterial] = 1.0827e23;
            }
            else
            {
               matCompare (material, "Nitride", &mode);
               if (mode >= matSUBCLASS)
               {
                  /* this is nitride; get the memory for the formula string */
                  if ((formula[iMaterial] =
		       (vChar *) vMalloc (sizeof (vChar) * 6)) == vNULL)
                  {
                     VeError3 (veSYS, "Could not allocate %s %d. %s",
                               "enough memory", iMaterial, "for formula");
                     VeReturn (vFALSE);
                  }
                  
                  /* set the formula */
                  vStrcpy (formula[iMaterial], "Si3N4");
                  density[iMaterial] = 9.0e22;
               }
               else
               {
                  matCompare (material, "Oxide", &mode);
                  if (mode >= matSUBCLASS)
                  {
                     /* this is oxide; get the memory for the formula string */
                     if ((formula[iMaterial] =
			  (vChar *) vMalloc (sizeof (vChar) * 5)) == vNULL)
                     {
                        VeError3 (veSYS, "Could not allocate %s %d. %s",
                                  "enough memory", iMaterial, "for formula");
			VeReturn (vFALSE);
                     }
                     
                     /* set the formula */
                     vStrcpy (formula[iMaterial], "SiO2");
                     density[iMaterial] = 7.0e22;
                  }
                  else
                  {
                     matCompare (material, "Si", &mode);
                     if (mode >= matSUBCLASS)
                     {
                        /* this is Silicon; get the memory for the formula
			   string */
                        if ((formula[iMaterial] =
			     (vChar *) vMalloc (sizeof (vChar) * 3)) ==
			    vNULL)
                        {
                           VeError3 (veSYS, "Could not allocate %s %d. %s",
                                     "enough memory", iMaterial,
				     "for formula");
                           VeReturn (vFALSE);
                        }
                        
                        /* set the formula */
                        vStrcpy (formula[iMaterial], "Si");
                        density[iMaterial] = 4.977e22;
                     }
		     else
                     {
                        matCompare (material, "GaAs", &mode);
                        if (mode >= matSUBCLASS)
                        {
                           /* this is Gallium Arsenide; get the memory for
                              the formula string */
                           if ((formula[iMaterial] =
                                (vChar *) vMalloc (sizeof (vChar) * 5)) ==
                               vNULL)
                           {
                              VeError3 (veSYS, "Could not allocate %s %d. %s",
                                        "enough memory", iMaterial,
                                        "for formula");
                              VeReturn (vFALSE);
                           }
                           
                           /* set the formula */
                           vStrcpy (formula[iMaterial], "GaAs");
                           density[iMaterial] = 4.4e22;
                        }
                        else
                        {
                           /* just copy the material string */
                           if ((formula[iMaterial] =
                                (vChar *) vMalloc (sizeof (vChar) *
                                                   (paiStrlen (material) + 1)))
                               == vNULL)
                           {
                              VeError3 (veSYS, "Could not allocate %s %d. %s",
                                        "enough memory", iMaterial,
                                        "for formula");
                              VeReturn (vFALSE);
                           }
                           
                           /* set the formula */
                           vStrcpy (formula[iMaterial], material);
                           density[iMaterial] = -1.0;
                        }
                     }
                  }
               }
            }
	    vFree (material);
	 }
      }
   }
   
   /* count the number of atom species */
   for (iMaterial = 0;
	iMaterial < G2sNumberOfSegments(g2sGeometry);
	iMaterial++)
   {
      material = formula[iMaterial];
      
      /* parse the string for atoms */
      while (*material != '\0')
      {
	 if (isupper (*material))
	    numAtoms++;
	 material++;
      }
   }
   
   /* allocate the memory for the atoms and probability arrays */
   if ((iAtom = (vInt *) vMalloc (sizeof (vInt) * numAtoms)) == vNULL)
   {
      VeError (veSYS, "Could not allocate enough memory for atoms array.");
      VeReturn (vFALSE);
   }
   if ((iProbability = (vInt *) vMalloc (sizeof (vInt) * numAtoms *
					 G2sNumberOfFaces(g2sGeometry))) ==
       vNULL)
   {
      VeError (veSYS, "Could not allocate enough memory for atoms array.");
      VeReturn (vFALSE);
   }
   
   /* initialize the probability array */
   for (iMaterial = 0;
	iMaterial < numAtoms * G2sNumberOfFaces(g2sGeometry);
	iMaterial++)
      iProbability[iMaterial] = 0;
   
   /* fill in the atomic numbers, the probabilities and determine the number
      of different atom species */
   for (iMaterial = 0;
	iMaterial < G2sNumberOfSegments(g2sGeometry);
	iMaterial++)
   {
      material = formula[iMaterial];
      if (density[iMaterial] < 0.0)
      {
         /* try to get the density from the Database */
         matGetPropertyReal (material, "Density", &(density[iMaterial]),
                             "g/cm^-3", &mode);
         if (mode <= matNOPROP)
            /* no density found */
            density[iMaterial] = -1.0;
	 if (density[iMaterial] >= 0.0)
	 {
	    /* get the mass */
	    matGetPropertyReal (material, "Mass", &(atomMass), "g", &mode);
	    if (mode <= matNOPROP)
	       /* no mass found */
	       density[iMaterial] = -1.0;
	    else
	       density[iMaterial] *= 1.0 / atomMass;
	 }
      }
      
      while (*material != '\0')
      {
         /* check, if the character is valid at all */
         if (!(isupper (*material) || islower (*material) ||
               isdigit (*material)))
         {
            VeError2 (veSYS, "Unexpected character <%c> in Material %s.",
                      *material, formula[iMaterial]);
            VeReturn (vFALSE);
         }
         
         /* Atoms must start capitalized! */
         if (!isupper (*material))
         {
            VeError1 (veSYS, "Chemical formula of element must start %s",
                      "with an uppercase letter.");
            VeReturn (vFALSE);
         }
         /* the first letter of the */
         cAtom[0] = *material;
         cAtom[1] = '\0';
         
         /* the next letter */
         material++;
         
         /* check the next letter of the formula */
         if (*material != '\0')
         {
            if (islower (*material))
            {
               /* this atom has a two-letter abbreviation */
               cAtom[1] = *material;
               material++;
            }
            
            /* is there a composition number now? */
            number = 1;
            if (*material != '\0')
            {
               if (isdigit (*material))
                  /* Yeah! */
                  number = (vInt) atoi (material);
               
               /* advance to the next non-digit character */ 
               while (isdigit (*material))
                  material++;
            }
         }
         
         /* now get the atomic number of the element */
         matGetPropertyInteger (cAtom, "AtomicNumber", &atomNumber, &mode);
         if (mode <= matNOPROP)
         {
            VeError1 (veSYS, "No atomic number for element <%s>.", cAtom);
            VeReturn (vFALSE);
         }
         
         /* search the atom in the iAtom array */
         for (iSearch = 0; iSearch < numDifferentAtoms; iSearch++)
            if (iAtom[iSearch] == atomNumber)
               break;
         
         if (iSearch == numDifferentAtoms)
         {
            /* found a new atom */
            iAtom[iSearch] = atomNumber;
            numDifferentAtoms++;
         }
         
         /* set the appropriate probabillity */
         iProbability[iMaterial*numAtoms+iSearch] = number;
      }
   }
   
   /* allocate the memory for the mass of the target atoms */
   if ((mass = (vDouble *) vMalloc (sizeof (vDouble) * numDifferentAtoms)) ==
       vNULL)
   {
      VeError (veSYS, "Could not allocate enough memory for mass array.");
      VeReturn (vFALSE);
   }
   
   /* get the atomic mass of the target atoms */
   for (iMaterial = 0; iMaterial < numDifferentAtoms; iMaterial++)
   {
      /* set the element name */
      if (iAtom[iMaterial] < 10)
      {
         sprintf (elementName, "Element%1d", iAtom[iMaterial]);
	 elementName[8] = '\0';
      }
      else
      {
	 sprintf (elementName, "Element%2d", iAtom[iMaterial]);
	 elementName[9] = '\0';
      }
      
      matGetPropertyReal (elementName, "Mass", &(mass[iMaterial]), "amu",
			  &mode);
      if (mode <= matNOPROP)
      {
         /* no mass found */
	 VeError1 (veSYS, "No mass for Element %s found.", elementName);
	 VeReturn (vFALSE);
      }
   }
   
   /* write everything to the output file */
   fprintf (outputFile, "First Comment Line\n");
   fprintf (outputFile, "Second Comment Line\n");
   fprintf (outputFile, " &IONS   DOSE=,E0=,NION=,&END\n");
   fprintf (outputFile, " &ATOMS  N2=%d,Z2=", numDifferentAtoms);
   for (iMaterial = 0; iMaterial < numDifferentAtoms; iMaterial++)
      fprintf (outputFile, "%d,", iAtom[iMaterial]);
   fprintf (outputFile, "RM2=");
   for (iMaterial = 0; iMaterial < numDifferentAtoms; iMaterial++)
      fprintf (outputFile, "%lf,", mass[iMaterial]);
   fprintf (outputFile, "&END\n");
   fprintf (outputFile, " &LAYER  NLAYER=%d,XTAL=",
	    G2sNumberOfFaces(g2sGeometry));
   for (iMaterial = 0; iMaterial < G2sNumberOfFaces(g2sGeometry); iMaterial++)
      fprintf (outputFile, "%c,", xtal[segmentNumber[iMaterial]-1]);
   fprintf (outputFile, "DENS=");
   for (iMaterial = 0; iMaterial < G2sNumberOfFaces(g2sGeometry); iMaterial++)
      if (density[segmentNumber[iMaterial]-1] > 0.0)
	 fprintf (outputFile, "%lfE22,",
		  density[segmentNumber[iMaterial]-1]/1.0e22);
      else
	 fprintf (outputFile, "no %s-density,",
		  formula[segmentNumber[iMaterial]-1]);
   fprintf (outputFile, "\n         ");
   for (iMaterial = 0; iMaterial < G2sNumberOfFaces(g2sGeometry); iMaterial++)
   {
      fprintf (outputFile, "PROB(1,%d)=", iMaterial+1);
      for (iSearch = 0; iSearch < numDifferentAtoms; iSearch++)
	 fprintf (outputFile, "%d.0,",
		  iProbability[numAtoms*(segmentNumber[iMaterial]-1)+iSearch]);
   }
   fprintf (outputFile, "\n         CGFILE='GEO.DAT',&END\n");
   fprintf (outputFile, " &OUTPUT LHIS=T,NIONBK=1000,&END\n");
   
   /* free the memory */
   for (iMaterial = 0;
	iMaterial < G2sNumberOfSegments(g2sGeometry);
	iMaterial++)
      vFree (formula[iMaterial]);
   vFree (formula);
   vFree (mass);
   vFree (iProbability);
   vFree (iAtom);
   vFree (xtal);
   vFree (density);
   vFree (segmentNumber);
   
   VeReturn (vTRUE);
}

/*** m a i n
     main routine for the conversion from PIF to the required structure. ***/

vInt main (vInt  argc,          /* argument counter */
	   vConstString *argv)  /* argument vector  */
{
   vConstString          physicalName,		/* physical PIF file name */
                         logicalName,		/* logical PIF file name */
                         geometryName,		/* PIF geometry name */
                         outputName,		/* output file name */
                         inputDeckName;		/* input deck file name */
   vosFile               vosFileName;		/* VOS file name */
   paiObject             inputFileHandle;	/* PIF input file handle */
   g2sGeometry_t         g2sGeometry;		/* geometry */
   g2sGeometryNames_t    g2sGeometryNames;	/* names of PIF objects */
   g2sGeometryHandles_t  inputHandles;		/* the geometry handles */
   gasAttributes_t       gasAttributes;		/* the pif attributes */
   FILE                 *outputFile;		/* output file */
   
   VeStart ("main");
   
   /* set up the VISTA error system */
   veSetupStandardErrorHandling ();
   veSetErrorLevel(veGetErrorHandler(veSTANDARD_ERROR_HANDLER), 
                   veLEVEL_WARNING);
   
   /* initialize the VISTA command line interface */
   if (vrrInitCommandLineResource(argc, argv, commandParameter, stderr) ==
       vFALSE)
      vExit(1);
   
   /* get the input data from the command line */
   readCommandLine (&physicalName, &logicalName, &geometryName,
		    &outputName, &inputDeckName);
   
   /* initialize the PIF Application Interface */
   if (palInitialize() == paiFALSE)
      VeFatal (veSYS, "Could not initialize PIF Application Interface.");
   
   /* convert the input PIF file name to an operating system independent
      name */
   vosFileName = vosOs2Vospec(physicalName, vFALSE);
   if (vosExistFile(vosFileName) == vFALSE)
      VeFatal1(veSYS, "Physical PIF file %s does not exist.", physicalName);
   
   /* If no logical input PIF file is specified, check if there is only
      one logical PIF file within the physical PIF file.
      If this is true take this logical PIF file otherwise abort. */
   if (logicalName == vNULL)
      logicalName = getOnlyLogicalPif(physicalName);
   if (logicalName == vNULL)
      VeFatal3 (veSYS, "%s %s %s.", "Do not know which logical PIF file",
		"should be opened in", physicalName);
   
   /* open the PIF file */
   inputFileHandle = palOpenFile (vosFileName, logicalName, palOPEN_EXISTING); 
   if (inputFileHandle == 0)
      VeFatal2 (veSYS, "Can not open logical PIF file %s in physical file %s.",
		logicalName, physicalName); 
   
   /* read the geometry from the PIF file */
   printf ("\n Reading geometry from %s:%s...\n", physicalName, logicalName);
   if (g2sReadGeometry (inputFileHandle, &inputHandles, &g2sGeometryNames, 
			&g2sGeometry, "Angstrom", geometryName, vNULL, vNULL,
			vNULL, vNULL, vNULL, vNULL, vNULL, vNULL, vNULL,
			vNULL) == vFALSE)
   {
      palCloseFile (inputFileHandle);
      VeFatal2 (veSYS, "Could not read geometry from PIF file %s:%s.",
		physicalName, logicalName);
   }
   printf (" Reading geometry succeeded!\n");
   
   /* close the input pif file */
   palCloseFile (inputFileHandle);
   
   /* open the output file */
   if ((outputFile = fopen (outputName, "w")) == NULL)
      VeFatal1 (veSYS, "Can not open output file %s.", outputName); 
   
   /* convert the geometry information to a fromat suitable for the
      generation of the quadtree and write it to the output file. */
   printf ("\n Converting geometry and writing it to %s...\n", outputName);
   if (convertGeometry (&g2sGeometry, outputFile) == vFALSE)
   {
      fclose (outputFile);
      vExit(3);
   }
   printf (" Writing geometry succeeded!\n");
   
   /* close the output file */
   fclose (outputFile);
   
   /* open the PIF file again */
   inputFileHandle = palOpenFile (vosFileName, logicalName, palOPEN_EXISTING); 
   if (inputFileHandle == 0)
      VeFatal2 (veSYS, "Can not open logical PIF file %s in physical file %s.",
		logicalName, physicalName); 
   
   /* free the operating system independent name */
   vosFreeVospec(vosFileName);
   
   /* allocate the attribute tree */
   printf ("\n Allocating attribute tree...\n");
   if (gasAllocateAttributeTree (&gasAttributes, inputFileHandle,
				 &g2sGeometry, "pif2imsil_", vFALSE,
				 (vConstString(*)())vNULL,
				 (vConstString(*)())vNULL,
				 (vBoolean(*)())vNULL, (paiObject(*)())vNULL,
				 (paiObject(*)())vNULL, (vLong(*)())vNULL,
				 (vLong(*)())vNULL) == vFALSE)
   {
      palCloseFile (inputFileHandle);
      VeFatal (veSYS, "Could not allocate attribute tree.");
   }
   printf (" Allocation succeeded!\n");
   
   /* read the attributes from the PIF file */
   printf ("\n Reading SegmentDescriptions from %s:%s...\n",
	   physicalName, logicalName);
   if (gasScanMultipleAttributes (vNULL, &gasAttributes, vNULL, vNULL, "",
				  "SegmentDescription", "*", "", vTRUE,
				  vTRUE, &inputHandles) == vFALSE)
   {
      palCloseFile (inputFileHandle);
      VeFatal3 (veSYS, "Could not read SegmentDescriptions %s %s:%s.",
		"from PIF file", physicalName, logicalName);
   }
   printf (" Reading SegmentDescriptions succeeded!\n");
   printf (" Reading MaterialTypes from %s:%s...\n",
	   physicalName, logicalName);
   if (gasScanMultipleAttributes (vNULL, &gasAttributes, "",
				  "SegmentDescription", "", "MaterialType",
				  "*", "", vTRUE, vTRUE, &inputHandles) ==
       vFALSE)
   {
      palCloseFile (inputFileHandle);
      VeFatal3 (veSYS, "Could not read MaterialTypes %s %s:%s.",
		"from PIF file", physicalName, logicalName);
   }
   printf (" Reading MaterialTypes succeeded!\n");
   
   /* open the VISTA material database */
   if (matOpen(paiNULL) == paiFALSE)
      VeFatal(veSYS, "Could not open vista material database.");
   
   /* open the input deck file */
   if ((outputFile = fopen (inputDeckName, "w")) == NULL)
   {
      matClose ();
      VeFatal1 (veSYS, "Can not open input deck file %s.", inputDeckName); 
   }
   
   /* convert the attribute information to the input deck fromat and
      generate the frame input deck file. */
   printf ("\n Converting attributes and writing them to %s...\n",
	   inputDeckName);
   if (convertAttributes (&gasAttributes, &g2sGeometry, outputFile) == vFALSE)
   {
      matClose ();
      fclose (outputFile);
      vExit(4);
   }
   printf (" Converting attributes succeeded!\n");
   
   /* close the output file */
   fclose (outputFile);
   
   /* close the material data base */
   matClose ();
   
   /* free gas attributes */
   gasFreeAttributeTree (&gasAttributes, vTRUE);
   
   /* close the input pif file again */
   palCloseFile (inputFileHandle);
   
   /* Free g2s geometry */
   g2sFreeGeometry (&g2sGeometry);
   g2sFreeGeometryNames (&g2sGeometryNames);
   
   /* Print termination message */
   printf ("\n Program terminated normally.\n\n");
   
   vExit(0);
}
